package edu.georgetown.gucs.experiment;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;

import javax.swing.UIManager;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jfree.chart.ChartColor;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.xy.XYDataItem;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.util.ShapeUtilities;

import com.lowagie.text.DocumentException;
import com.lowagie.text.Rectangle;
import com.lowagie.text.pdf.DefaultFontMapper;
import com.lowagie.text.pdf.PdfContentByte;
import com.lowagie.text.pdf.PdfTemplate;
import com.lowagie.text.pdf.PdfWriter;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.tokenizers.Splitter;
import edu.georgetown.gucs.utility.BoundedExecutor;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectCloner;
import edu.georgetown.gucs.utility.Pair;

/**
 * Runs many {@linkplain Trial}s to determine the best settings to use for {@linkplain Dictionary} and
 * {@linkplain Fingerprinter} creation. Setting results print to standard output and optionally written to a database.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class Experiment {

  private boolean insert;
  private int numberOfTrials;
  private int numberOfTrialThreads;
  private int completedTrials = 0;
  private String language;
  private String description;
  private String config_file;
  private String dataset;
  private String manglerPlots = "";
  private String plotsPath = "plots/";
  private File datasetPath;
  private FileLister datasetFiles;
  private DBInterface database;
  private List<Integer> completedExperimentIDs = new Vector<Integer>();
  private List<Long> randomSeeds = new Vector<Long>();
  private Set<String> plots = new HashSet<String>();
  private Map<List<String>, List<String>> tokenizers_manglers = new HashMap<List<String>, List<String>>();
  private List<String> fingerprinters = new Vector<String>();
  private List<Double> dictionaryParameters = new Vector<Double>();
  private List<Double> sampleParameters = new Vector<Double>();
  private Vector<Pair<Double, Double>> IDFs = new Vector<Pair<Double, Double>>();
  private Vector<Pair<String, Integer>> matcherVector = new Vector<Pair<String, Integer>>();
  private Map<String, Double[]> manglerResultCounts = new HashMap<String, Double[]>();
  private List<Splitter> splitters = new Vector<Splitter>();
  /**
   * Map of TrialParameters to a LinkedHashMap of manglerName to vector of results (precision/recall/fscore/TP/FP/TN/FN)
   */
  private Map<TrialParameters, LinkedHashMap<String, Vector<Double[]>>> experimentManglerResults = new LinkedHashMap<TrialParameters, LinkedHashMap<String, Vector<Double[]>>>();
  private List<Double> precisionVector = new Vector<Double>();
  private List<Double> recallVector = new Vector<Double>();
  private List<Double> fscoreVector = new Vector<Double>();
  private List<Double> dictionarySizeVector = new Vector<Double>();

  private List<XYDataItem> precision_recall = new Vector<XYDataItem>();
  private List<XYDataItem> precision_fscore = new Vector<XYDataItem>();
  private List<XYDataItem> recall_fscore = new Vector<XYDataItem>();
  private List<XYDataItem> dictionary_precision = new Vector<XYDataItem>();
  private List<XYDataItem> dictionary_recall = new Vector<XYDataItem>();
  private List<XYDataItem> dictionary_fscore = new Vector<XYDataItem>();
  private List<XYDataItem> deleteToken_precision = new Vector<XYDataItem>();
  private List<XYDataItem> deleteToken_recall = new Vector<XYDataItem>();
  private List<XYDataItem> deleteChar_precision = new Vector<XYDataItem>();
  private List<XYDataItem> deleteChar_recall = new Vector<XYDataItem>();
  private List<XYDataItem> deleteWhitespace_precision = new Vector<XYDataItem>();
  private List<XYDataItem> deleteWhitespace_recall = new Vector<XYDataItem>();
  private List<XYDataItem> changeToChar_precision = new Vector<XYDataItem>();
  private List<XYDataItem> changeToChar_recall = new Vector<XYDataItem>();
  private List<XYDataItem> changeToken_precision = new Vector<XYDataItem>();
  private List<XYDataItem> changeToken_recall = new Vector<XYDataItem>();
  private List<XYDataItem> splitToken_precision = new Vector<XYDataItem>();
  private List<XYDataItem> splitToken_recall = new Vector<XYDataItem>();
  private List<XYDataItem> changeToPunc_precision = new Vector<XYDataItem>();
  private List<XYDataItem> changeToPunc_recall = new Vector<XYDataItem>();
  private List<XYDataItem> deleteSection_precision = new Vector<XYDataItem>();
  private List<XYDataItem> deleteSection_recall = new Vector<XYDataItem>();
  private List<XYDataItem> deleteHead_precision = new Vector<XYDataItem>();
  private List<XYDataItem> deleteHead_recall = new Vector<XYDataItem>();
  private List<XYDataItem> deleteTail_precision = new Vector<XYDataItem>();
  private List<XYDataItem> deleteTail_recall = new Vector<XYDataItem>();
  private List<XYDataItem> summaryReduce_precision = new Vector<XYDataItem>();
  private List<XYDataItem> summaryReduce_recall = new Vector<XYDataItem>();

  private final static int BLACK_WHITE = 100;
  private final static int COLOR = 101;
  private final static Color[] GRAY_SCALE = new Color[] { Color.BLACK, Color.GRAY, Color.DARK_GRAY };
  private final static Color[] DARK_COLORS = new Color[] { Color.DARK_GRAY, ChartColor.DARK_RED, ChartColor.DARK_BLUE,
      ChartColor.DARK_GREEN, ChartColor.DARK_MAGENTA, ChartColor.DARK_CYAN, };
  private final static Shape[] SHAPES = new Shape[] { ShapeUtilities.createRegularCross(3, 1),
      ShapeUtilities.createDiagonalCross(3, 1), ShapeUtilities.createUpTriangle(3),
      ShapeUtilities.createDownTriangle(3), ShapeUtilities.createDiamond(3) };

  /**
   * Constructor that loads properties from XML configuration file
   * 
   * @param config_file
   *          the string XML filename that specifies the configurations to be used
   */
  public Experiment(String config_file) {
    Global.monitor.writeToBuffer("Experiment", "Constructor", "config_file=" + config_file);
    this.config_file = config_file;
    loadXMLConfig();
  }

  /**
   * Load properties from XML configuration file
   */
  @SuppressWarnings("unchecked")
  private void loadXMLConfig() {
    try {
      File xml = new File(this.config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);

      Element root = doc.getRootElement();
      String threads = root.getAttributeValue("threads");
      this.numberOfTrialThreads = (threads != null && threads.length() > 0) ? Integer.valueOf(threads) : Runtime
          .getRuntime().availableProcessors();
      Global.monitor.writeToBuffer("Experiment", "loadXMLConfig", "numberOfTrialThreads=" + this.numberOfTrialThreads);

      String trials = root.getAttributeValue("trials");
      this.numberOfTrials = (trials == null || Integer.valueOf(trials) < 3) ? 3 : Integer.valueOf(trials);
      Global.monitor.writeToBuffer("Experiment", "loadXMLConfig", "numberOfTrials=" + this.numberOfTrials);

      String dbInsert = root.getAttributeValue("DB_insert");
      this.insert = (dbInsert != null) ? Boolean.valueOf(dbInsert) : false;
      Global.monitor.writeToBuffer("Experiment", "loadXMLConfig", "insert=" + this.insert);

      if ((this.description = root.getAttributeValue("description")) == null) {
        Global.monitor.printErrorMessage("Experiment", "WARNING: Null description");
        Global.monitor.writeToBuffer("Experiment", "loadXMLConfig", "WARNING: Null description");
      }

      if (root.indexOf(root.getChild("plots")) != -1) {
        this.plotsPath = root.getChild("plots").getAttributeValue("path");
        for (Element plot : (List<Element>) root.getChild("plots").getChildren("plot")) {
          this.plots.add(plot.getAttributeValue("name").trim().toLowerCase(Locale.ENGLISH));
          if (plot.getAttributeValue("name").equalsIgnoreCase("mangler")) {
            this.manglerPlots += (plot.getAttributeValue("parameter") == null) ? "" : plot.getAttributeValue(
                "parameter").toLowerCase(Locale.ENGLISH)
                + " ";
          }
        }
      }
      for (Element seed : (List<Element>) root.getChildren("randomSeed")) {
        this.randomSeeds.add(Long.valueOf(seed.getAttributeValue("seed")));
      }
      if (this.randomSeeds.size() != 0 && this.numberOfTrials != this.randomSeeds.size()) {
        ExceptionHandler.handleExceptionThrow(new Exception(
            "Experiment: Number of trials does not match number of random seeds. " + "Trials: " + this.numberOfTrials
                + " seeds size: " + this.randomSeeds.size()));
      }

      Element data = root.getChild("dataset");
      this.dataset = data.getAttributeValue("name");
      if (this.dataset == null || this.dataset.length() == 0) {
        ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No dataset name provided"));
      }
      Global.monitor.writeToBuffer("Experiment", "loadXMLConfig", "dataset=" + this.dataset);

      String path = data.getAttributeValue("path");
      if (path == null || path.length() == 0) {
        ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No path to dataset provided"));
      }
      this.datasetPath = new File(path);
      Global.monitor.writeToBuffer("Experiment", "loadXMLConfig", "datasetPath=" + this.datasetPath);

      for (Element sampleParameter : (List<Element>) data.getChildren("sample")) {
        if (sampleParameter.getAttributeValue("count") != null) {
          double parameter = Double.valueOf(sampleParameter.getAttributeValue("count"));
          if (parameter < 0) {
            ExceptionHandler.handleExceptionThrow(new Exception("Experiment: Negative sample counts"));
          } else if (parameter > 5000) {
            Global.monitor.printErrorMessage("Experiment",
                "Sample counts greater than 5000 can take a really long time when run with many manglers");
            Global.monitor.writeToBuffer("Experiment", "loadXMLConfig",
                "Sample counts greater than 5000 can take a really long time when run with many manglers");
          }
          this.sampleParameters.add(parameter);
        }
        if (sampleParameter.getAttributeValue("percent") != null) {
          double parameter = Double.valueOf(sampleParameter.getAttributeValue("percent"));
          if (parameter < 1 || parameter > 100) {
            ExceptionHandler.handleExceptionThrow(new Exception("Experiment: Sample percentage out of range"));
          }
          this.sampleParameters.add(parameter / 1000);
        }
      }
      if (this.sampleParameters == null || this.sampleParameters.size() == 0) {
        ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No sample settings provided"));
      }

      Element dict = data.getChild("dictionary");
      String lang = dict.getAttributeValue("language");
      this.language = (lang != null) ? lang : "english";
      for (Element dictParameter : (List<Element>) dict.getChildren("create")) {
        if (dictParameter.getAttributeValue("count") != null) {
          double parameter = Double.valueOf(dictParameter.getAttributeValue("count"));
          if (parameter < 0) {
            ExceptionHandler.handleExceptionThrow(new Exception("Experiment: Negative dictionary counts"));
          } else if (parameter == 0) {
            ExceptionHandler.handleExceptionThrow(new Exception("Experiment: Zero dictionary counts"));
          }
          this.dictionaryParameters.add(parameter);

        }
        if (dictParameter.getAttributeValue("percent") != null) {
          double parameter = Double.valueOf(dictParameter.getAttributeValue("percent"));
          if (parameter < 1 || parameter > 100) {
            ExceptionHandler.handleExceptionThrow(new Exception("Experiment: Dictionary percentage out of range"));
          }
          this.dictionaryParameters.add(parameter / 1000);
        }
      }
      if (this.dictionaryParameters == null || this.dictionaryParameters.size() == 0) {
        ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No dictionary settings provided"));
      }

      if (dict.getChild("IDF") != null) {
        for (Element idf : (List<Element>) dict.getChildren("IDF")) {
          Double minIDF = Double.valueOf(idf.getAttributeValue("min"));
          Double maxIDF = Double.valueOf(idf.getAttributeValue("max"));
          if (minIDF < maxIDF) {
            this.IDFs.add(new Pair<Double, Double>(minIDF, maxIDF));
          } else {
            ExceptionHandler.handleExceptionThrow(new Exception("Experiment: The minimum IDF value (" + minIDF
                + ") is >= the maximum IDF value (" + maxIDF + ")"));
          }
        }
      } else {
        this.IDFs.add(new Pair<Double, Double>(0.0, 1.0));
      }

      for (Element fingerprint : (List<Element>) root.getChildren("fingerprinter")) {
        this.fingerprinters.add(fingerprint.getAttributeValue("name"));
      }
      if (this.fingerprinters == null || this.fingerprinters.size() == 0) {
        ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No fingerprinter provided"));
      }

      for (Element matcherElement : (List<Element>) root.getChildren("matcher")) {
        if (matcherElement.getAttributeValue("score") != null) {
          this.matcherVector.add(new Pair<String, Integer>(matcherElement.getAttributeValue("name"), Integer
              .valueOf(matcherElement.getAttributeValue("score"))));
        } else {
          // If there are no matcher parameters then add the default minimum score (0) to the vector
          this.matcherVector.add(new Pair<String, Integer>(matcherElement.getAttributeValue("name"), 0));
        }
      }
      if (this.matcherVector == null || this.matcherVector.isEmpty()) {
        ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No matcher provided"));
      }

      loadTokenizers(root, "splitters");
      loadTokenizers(root, "tokenizers");

    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  } // end loadXMLConfig method

  /**
   * Loads the tokenizers and splitters given in the XML configuration file for this experiment
   * 
   * @param root
   *          the root element from the XML configuration file for this experiment
   * @param name
   *          the string name of the element to load (either "splitters" or "tokenizers")
   */
  @SuppressWarnings("unchecked")
  private void loadTokenizers(Element root, String name) {
    for (Element element : (List<Element>) root.getChildren(name)) {
      Vector<String> tokenizerNames = new Vector<String>();
      Vector<String> manglerLists = new Vector<String>();
      for (Element tokenizerElement : (List<Element>) element.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizerNames.size() == 0 && !tokenizer.contains("FileTokenizer")) {
          ExceptionHandler.handleExceptionThrow(new Exception("Experiment: First tokenizer (" + tokenizer
              + ") must read from a file."));
        }

        if (tokenizerElement.getAttributeValue("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
        } else if (tokenizer.equalsIgnoreCase("FileManglerTokenizer")) {
          for (Element manglersListElement : (List<Element>) tokenizerElement.getChildren("manglers")) {
            StringBuffer mangler = new StringBuffer();
            for (Element manglerElement : (List<Element>) manglersListElement.getChildren("mangler")) {
              mangler.append(" " + manglerElement.getAttributeValue("name").trim());
              if (!manglerElement.getAttributeValue("name").equalsIgnoreCase("null")) {
                mangler.append("-" + manglerElement.getAttributeValue("percent").trim());
              }
            }
            manglerLists.add(mangler.toString().trim());
          }
          manglerLists.trimToSize();
          if (manglerLists.size() == 0) {
            ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No manglers provided"));
          }
        }
        tokenizerNames.add(tokenizer.trim());
      }
      tokenizerNames.trimToSize();

      for (Element splittersElement : (List<Element>) element.getChildren("splitters")) {
        this.splitters.add(new Splitter(tokenizerNames, splittersElement));
      }

      if (name.equalsIgnoreCase("splitters")) {
        // TODO: !!! only add if "tokenizers"?
        // this.tokenizers_manglers.put(this.splitters., manglerLists);
      } else {
        this.tokenizers_manglers.put(tokenizerNames, manglerLists);
      }
    }
    if ((this.tokenizers_manglers == null || this.tokenizers_manglers.size() == 0)
        && !name.equalsIgnoreCase("splitters")) {
      ExceptionHandler.handleExceptionThrow(new Exception("Experiment: No tokenizers provided"));
    }
  }

  /**
   * Prints out the properties for this experiment
   */
  public void showSettings() {
    int manglerCount = 0;
    System.out.println("Dataset: " + this.dataset);
    System.out.println("Data path: " + this.datasetPath.getAbsolutePath());
    System.out.println("Description: " + this.description);
    System.out.println("Language: " + this.language);
    System.out.println("Database Insert: " + this.insert);
    System.out.println("Number of threads: " + this.numberOfTrialThreads);
    System.out.println("Number of trials: " + this.numberOfTrials);
    System.out.println("Log file: " + Global.monitor.getOutputFile());
    System.out.println();
    if (this.randomSeeds.size() > 0) {
      System.out.println("Random Seeds:");
      for (Long seed : this.randomSeeds) {
        System.out.println("     " + seed.toString());
      }
      System.out.println();
    }
    System.out.println("Dictionary settings:");
    for (Double parameter : this.dictionaryParameters) {
      if (parameter < 1) {
        System.out.println(" Percents: " + (int) (parameter * 1000));
      } else {
        System.out.println(" File Count: " + parameter.intValue());
      }
    }
    System.out.println();
    System.out.println("Sample settings:");
    for (Double parameter : this.sampleParameters) {
      if (parameter < 1) {
        System.out.println(" Percents: " + (int) (parameter * 1000));
      } else {
        System.out.println(" File Count: " + parameter.intValue());
      }
    }
    System.out.println();
    System.out.format("%-10s %-10s %n", "Min IDFs", "Max IDFs");
    for (Pair<Double, Double> idfRange : this.IDFs) {
      System.out.format(" %-10s %-10s %n", idfRange.getA(), idfRange.getB());
    }
    System.out.println();
    System.out.println("Tokenizer Lists:");
    for (List<String> tokenizerList : this.tokenizers_manglers.keySet()) {
      System.out.println(" Tokenizers: " + tokenizerList);
      if (tokenizerList.contains("FileManglerTokenizer")) {
        System.out.println("   Manglers: " + this.tokenizers_manglers.get(tokenizerList));
        manglerCount += this.tokenizers_manglers.get(tokenizerList).size();
      }
    }
    System.out.println();
    if (this.splitters.size() > 0) {
      System.out.println("Splitter List:");
      for (Splitter splitter : this.splitters) {
        System.out.println("  Splitters: " + splitter.getSplitterNames());
        System.out.println(" Tokenizers: " + splitter.getTokenizersNames());
        System.out.println();
      }

      System.out.println();
    }
    System.out.println("Fingerprinter: ");
    for (String fingerprinter : this.fingerprinters) {
      System.out.println("     " + fingerprinter);
    }
    System.out.println("Matcher: ");
    for (Pair<String, Integer> matcherPair : this.matcherVector) {
      System.out.println("     " + matcherPair.getA() + "(" + matcherPair.getB() + ")");
    }
    System.out.println();
    if (this.plots.size() > 0) {
      System.out.println("Plots: " + this.plotsPath);
      for (String plot : this.plots) {
        System.out.print("     " + plot);
        if (plot.equalsIgnoreCase("mangler")) {
          System.out.print(":  " + this.manglerPlots);
        }
        System.out.println();
      }
      System.out.println();
    }
    int splitterCount = (this.splitters.size() > 0) ? this.splitters.size() : 1;
    int totalExperiments = this.numberOfTrials * this.dictionaryParameters.size() * this.sampleParameters.size()
        * this.IDFs.size() * splitterCount * this.tokenizers_manglers.keySet().size() * manglerCount
        * this.fingerprinters.size() * this.matcherVector.size();
    System.out.println("Total runs: " + totalExperiments);
  } // end showSettings method

  /**
   * Runs this experiment
   * 
   * @return the <code>int</code> database run number for this experiment; returns 0 if not inserting into the database
   */
  public int runExperiment() {
    Global.monitor.printStartMessage("Experiment", "Experiment start");
    if (this.insert) {
      // Connect to the correct database
      this.database = new DBInterface(this.dataset);
      Global.monitor.printMessage("Experiment", "Connected to " + this.dataset + " database");
      Global.monitor.writeToBuffer("Experiment", "runExperiment", "Connected to " + this.dataset + " database");
    }

    Global.monitor.printMessage("Experiment", "Running Trials");
    int runNumber = runTrials();

    Global.monitor.printMessage("Experiment", "Printing Trial Results");
    printTrialResults();
    printExperimentResults();

    if (this.plots.size() > 0) {
      Global.monitor.printMessage("Experiment", "Creating plots: dataset=" + this.dataset + "  runNumber=" + runNumber);
      if (this.plots.contains("dictionary_precision") || this.plots.contains("dictionary_recall")
          || this.plots.contains("dictionary_fscore") || this.plots.contains("dictionary_precision_recall")) {
        populateDictionaryResults();
      }
      if (!this.manglerPlots.isEmpty()) {
        populateManglerResults();
      }
      if (runNumber > 0) {
        createPlots(this.plotsPath + this.dataset + runNumber, COLOR);
      } else
        createPlots(this.plotsPath + this.dataset, COLOR);
    }

    if (this.insert) {
      Global.monitor.printMessage("Experiment", "Computing and storing experiment results in database");
      Global.monitor.writeToBuffer("Experiment", "runExperiment",
          "Computing and storing experiment results in database");
      for (Integer experimentID : this.completedExperimentIDs) {
        this.database.insertExperimentResults(experimentID, this.numberOfTrials);
      }
      this.database.finishRun(runNumber);
      this.database.close();
      this.database = null;
    }

    Global.monitor.printStopMessage("Experiment", "Experiment complete");
    Global.monitor.writeToBuffer("Experiment", "runExperiment", "Experiment Complete");
    return runNumber;
  }

  /**
   * Runs trials for each setting
   * 
   * @return the <code>int</code> database run number for this experiment; returns 0 if not inserting into the database
   */
  public int runTrials() {
    Global.monitor.writeToBuffer("Experiment", "runTrials", "start");
    int runNumber = (this.insert) ? this.database.startNewRun() : 0;

    // Get the listing of files from the dataset. This is slow, so we only want to do it once
    this.datasetFiles = new FileLister(this.datasetPath.getAbsolutePath(), this.numberOfTrialThreads);

    // Create a thread pool to use to run the trials
    ExecutorService exec = Executors.newFixedThreadPool(this.numberOfTrialThreads);
    BoundedExecutor bex = new BoundedExecutor(exec, this.numberOfTrialThreads);

    for (Double currentDictionaryParameter : this.dictionaryParameters) {

      for (Double currentSampleParameter : this.sampleParameters) {

        Global.monitor.printMessage("Experiment", "ObjectCloner.deepCopy(" + this.datasetFiles.getDirectory() + ")");
        FileLister currentFileLister = (FileLister) ObjectCloner.deepCopy(this.datasetFiles);
        Global.monitor.printMessage("Experiment", "FileLister createSublists");
        currentFileLister.createSublists(currentDictionaryParameter, currentSampleParameter);

        for (Entry<List<String>, List<String>> currentTokenizers : this.tokenizers_manglers.entrySet()) {

          // TODO: !!! figure out how to run tokenizers and splitters in experiment

          Global.monitor.printMessage("Experiment", "Creating Dictionary (" + currentFileLister.dictionary_size()
              + " files)");
          Dictionary dictionary = new Dictionary();
          dictionary.threadingOn(this.numberOfTrialThreads);
          dictionary.setLocale(this.language);
          dictionary.setTokenizers(currentTokenizers.getKey());
          dictionary.createPartialDictionary(currentFileLister);
          Global.monitor.printMessage("Experiment",
              "Finished Partial Dictionary Creation (" + dictionary.getTotalDocuments() + " total documents)");

          for (Pair<Double, Double> currentIDF : this.IDFs) {

            Dictionary currentDictionary = new Dictionary();
            currentDictionary.clone(dictionary);
            // (Dictionary) ObjectCloner.deepCopy(dictionary);

            if ((currentIDF.getA() > 0) || (currentIDF.getB() < 1)) {
              Global.monitor.printMessage("Experiment", "Trimming dictionary: min=" + currentIDF.getA() + " max="
                  + currentIDF.getB());
              currentDictionary.trimByIDF(currentIDF.getA(), currentIDF.getB());
            }

            for (String currentFingerprinterName : this.fingerprinters) {

              for (Pair<String, Integer> currentMatcherPair : this.matcherVector) {

                for (String manglers : currentTokenizers.getValue()) {
                  Global.monitor.printMessage("Experiment", "Mangler(s): " + manglers);

                  int experimentID = -1;
                  if (this.insert) {
                    experimentID = this.database.newExperiment(runNumber, this.description, this.dataset,
                        this.datasetPath.getAbsolutePath(), this.config_file, currentFingerprinterName,
                        currentMatcherPair.getA(), currentMatcherPair.getB(), currentIDF.getA(), currentIDF.getB(),
                        currentSampleParameter, currentDictionaryParameter, this.language);
                    this.database.saveTokenizers(experimentID, currentTokenizers.getKey());
                    this.completedExperimentIDs.add(experimentID);
                  }

                  // Run the trials
                  TrialParameters trialParameters = null;
                  LinkedHashMap<String, Vector<Double[]>> manglerResults = new LinkedHashMap<String, Vector<Double[]>>();
                  for (int i = 0; i < this.numberOfTrials; i++) {
                    // Create a random seed, if there isn't one
                    long currentSeed = (this.randomSeeds == null || this.randomSeeds.size() == 0) ? System
                        .currentTimeMillis() : this.randomSeeds.get(i);

                    Global.monitor.printMessage("Experiment", "Trial #" + (i + 1));

                    Trial trial = new Trial(experimentID, this.insert, this.database, this.dataset,
                        currentFileLister.getSampleFiles(), currentSeed, currentDictionaryParameter,
                        currentSampleParameter, currentDictionary, manglers, currentFingerprinterName,
                        currentMatcherPair.getA(), currentMatcherPair.getB());

                    try {
                      bex.submitTask(trial);
                    } catch (RejectedExecutionException e) {
                      ExceptionHandler.handleExceptionExit(e);
                    }

                    Map<String, Double[]> trialManglerResults = trial.getManglerResults();
                    Global.monitor.printMessage("Experiment", "Retrieved Trial Results");
                    trialParameters = trial.getTrialParameters();
                    this.completedTrials++;

                    for (Entry<String, Double[]> mangler_results : trialManglerResults.entrySet()) {
                      Double[] runResult = new Double[] { mangler_results.getValue()[Global.PRECISION],
                          mangler_results.getValue()[Global.RECALL], mangler_results.getValue()[Global.FSCORE],
                          mangler_results.getValue()[Global.TRUE_POSITIVE],
                          mangler_results.getValue()[Global.FALSE_POSITIVE],
                          mangler_results.getValue()[Global.TRUE_NEGATIVE],
                          mangler_results.getValue()[Global.FALSE_NEGATIVE],
                          mangler_results.getValue()[Global.DICTIONARY_SIZE] };
                      this.dictionarySizeVector.add(mangler_results.getValue()[Global.DICTIONARY_SIZE]);

                      // Record all precision, recall, and f-score for total averages (over the experiment)
                      this.precisionVector.add(runResult[Global.PRECISION]);
                      this.recallVector.add(runResult[Global.RECALL]);
                      this.fscoreVector.add(runResult[Global.FSCORE]);
                      this.precision_recall.add(new XYDataItem(runResult[Global.PRECISION], runResult[Global.RECALL]));
                      this.precision_fscore.add(new XYDataItem(runResult[Global.PRECISION], runResult[Global.FSCORE]));
                      this.recall_fscore.add(new XYDataItem(runResult[Global.RECALL], runResult[Global.FSCORE]));

                      // Keep track of the counts for each of the different manglers
                      if (this.manglerResultCounts.containsKey(mangler_results.getKey())) {
                        Double[] oldCounts = this.manglerResultCounts.get(mangler_results.getKey());
                        this.manglerResultCounts.put(mangler_results.getKey(), new Double[] {
                            oldCounts[Global.PRECISION] + mangler_results.getValue()[Global.PRECISION],
                            oldCounts[Global.RECALL] + mangler_results.getValue()[Global.RECALL],
                            oldCounts[Global.FSCORE] + mangler_results.getValue()[Global.FSCORE],
                            oldCounts[Global.COUNT] + 1 });
                      } else {
                        this.manglerResultCounts.put(mangler_results.getKey(),
                            new Double[] { mangler_results.getValue()[Global.PRECISION],
                                mangler_results.getValue()[Global.RECALL], mangler_results.getValue()[Global.FSCORE],
                                1.0 });
                      }

                      Vector<Double[]> trialResults = (i == 0) ? new Vector<Double[]>() : manglerResults
                          .get(mangler_results.getKey());
                      trialResults.add(runResult);
                      manglerResults.put(mangler_results.getKey(), trialResults);
                      Global.monitor.printMessage("Experiment", "Trial Results Recorded: " + mangler_results.getKey());
                    }

                  }
                  Global.monitor.printMessage("", "");
                  this.experimentManglerResults.put(trialParameters, manglerResults);
                }
              }
            }
          }
        }
      }
    }

    exec.shutdown();

    while (!exec.isTerminated()) {
      // Wait a second and see if we are done
      try {
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    Global.monitor.writeToBuffer("Experiment", "runTrials", "end");
    return runNumber;
  } // runTrial

  /**
   * Calculates the results of all trials, grouped by parameters and displayed by mangler
   */
  public void printTrialResults() {
    System.out.println();
    System.out.println("**************");
    System.out.println("Trial Results");
    System.out.println("**************");
    for (TrialParameters trialParameters : this.experimentManglerResults.keySet()) {
      System.out.println(trialParameters);
      System.out.println();
      for (String mangler : this.experimentManglerResults.get(trialParameters).keySet()) {
        System.out.format(" %-15s %s%n", "Mangler(s):", mangler);
        double[] resultSum = new double[] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
        for (Double[] manglerResults : this.experimentManglerResults.get(trialParameters).get(mangler)) {
          // Add results to sum array of result type (to be used to find average over trials)
          resultSum[Global.PRECISION] += manglerResults[Global.PRECISION];
          resultSum[Global.RECALL] += manglerResults[Global.RECALL];
          resultSum[Global.FSCORE] += manglerResults[Global.FSCORE];
          resultSum[Global.TRUE_POSITIVE] += manglerResults[Global.TRUE_POSITIVE];
          resultSum[Global.FALSE_POSITIVE] += manglerResults[Global.FALSE_POSITIVE];
          resultSum[Global.TRUE_NEGATIVE] += manglerResults[Global.TRUE_NEGATIVE];
          resultSum[Global.FALSE_NEGATIVE] += manglerResults[Global.FALSE_NEGATIVE];
          resultSum[Global.DICTIONARY_SIZE] += manglerResults[Global.DICTIONARY_SIZE];

          // Print mangler results by result type
          System.out.format("%13s: %-15.6f", "Precision", manglerResults[Global.PRECISION]);
          System.out.format("%13s: %-15.6f", "Recall", manglerResults[Global.RECALL]);
          System.out.format("%13s: %-15.6f", "F-Score", manglerResults[Global.FSCORE]);
          System.out.format("%6s: %-15d", "TP", manglerResults[Global.TRUE_POSITIVE].intValue());
          System.out.format("%6s: %-15d", "FP", manglerResults[Global.FALSE_POSITIVE].intValue());
          System.out.format("%6s: %-15d", "TN", manglerResults[Global.TRUE_NEGATIVE].intValue());
          System.out.format("%6s: %-15d", "FN", manglerResults[Global.FALSE_NEGATIVE].intValue());
          System.out.println();
        }

        // Calculate averages for each result type
        int numManglerResults = this.experimentManglerResults.get(trialParameters).get(mangler).size();
        Double[] averages = new Double[] { resultSum[Global.PRECISION] / numManglerResults,
            resultSum[Global.RECALL] / numManglerResults, resultSum[Global.FSCORE] / numManglerResults,
            resultSum[Global.TRUE_POSITIVE] / numManglerResults, resultSum[Global.FALSE_POSITIVE] / numManglerResults,
            resultSum[Global.TRUE_NEGATIVE] / numManglerResults, resultSum[Global.FALSE_NEGATIVE] / numManglerResults,
            resultSum[Global.DICTIONARY_SIZE] / numManglerResults };
        Vector<Double[]> averageVector = this.experimentManglerResults.get(trialParameters).get(mangler);
        averageVector.add(0, averages);
        this.experimentManglerResults.get(trialParameters).put(mangler, averageVector);

        // Print mangler result averages by result type
        System.out.format("%13s: %-15.6f", "AvgPrecision", averages[Global.PRECISION]);
        System.out.format("%13s: %-15.6f", "AvgRecall", averages[Global.RECALL]);
        System.out.format("%13s: %-15.6f", "AvgF-Score", averages[Global.FSCORE]);
        System.out.format("%6s: %-15.6f", "AvgTP", averages[Global.TRUE_POSITIVE]);
        System.out.format("%6s: %-15.6f", "AvgFP", averages[Global.FALSE_POSITIVE]);
        System.out.format("%6s: %-15.6f", "AvgTN", averages[Global.TRUE_NEGATIVE]);
        System.out.format("%6s: %-15.6f", "AvgFN", averages[Global.FALSE_NEGATIVE]);
        System.out.println();
        System.out.println();
      }
    }
  }

  /**
   * Prints the results of all trials, grouped by parameters and displayed by mangler
   */
  public void printExperimentResults() {
    System.out.println("*******************");
    System.out.println("Experiment Results");
    System.out.println("*******************");
    final double tValue = Statistics.getTValueNinetyFive(this.completedTrials - 1);

    System.out.println("Precision:");
    double mean = Statistics.getMean(this.precisionVector);
    double stdDev = Statistics.getStandardDeviation(this.precisionVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(this.completedTrials)) * tValue);
    System.out.println("");

    System.out.println("Recall:");
    mean = Statistics.getMean(this.recallVector);
    stdDev = Statistics.getStandardDeviation(this.recallVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(this.completedTrials)) * tValue);
    System.out.println("");

    System.out.println("F-Score:");
    mean = Statistics.getMean(this.fscoreVector);
    stdDev = Statistics.getStandardDeviation(this.fscoreVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(this.completedTrials)) * tValue);
    System.out.println("");

    System.out.println("Dictionary Size:");
    mean = Statistics.getMean(this.dictionarySizeVector);
    stdDev = Statistics.getStandardDeviation(this.dictionarySizeVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(this.completedTrials)) * tValue);
    System.out.println("");
  }

  /**
   * Populates the arrays of points for the different dictionary size and relevance measures
   */
  private void populateDictionaryResults() {
    Global.monitor.writeToBuffer("Experiment", "populateDictionaryResults", "start");
    for (Entry<TrialParameters, LinkedHashMap<String, Vector<Double[]>>> trialParameters : this.experimentManglerResults
        .entrySet()) {
      for (Entry<String, Vector<Double[]>> mangler : this.experimentManglerResults.get(trialParameters.getKey())
          .entrySet()) {

        if (this.plots.contains("dictionary_precision") || this.plots.contains("dictionary_precision_recall")) {
          this.dictionary_precision.add(new XYDataItem((double) mangler.getValue().get(0)[Global.DICTIONARY_SIZE],
              (mangler.getValue().get(0)[Global.PRECISION] / this.numberOfTrials)));
        }
        if (this.plots.contains("dictionary_recall") || this.plots.contains("dictionary_precision_recall")) {
          this.dictionary_recall.add(new XYDataItem((double) mangler.getValue().get(0)[Global.DICTIONARY_SIZE],
              (mangler.getValue().get(0)[Global.RECALL] / this.numberOfTrials)));
        }
        if (this.plots.contains("dictionary_fscore")) {
          this.dictionary_fscore.add(new XYDataItem((double) mangler.getValue().get(0)[Global.DICTIONARY_SIZE],
              (mangler.getValue().get(0)[Global.FSCORE] / this.numberOfTrials)));
        }
      }
    }
    Global.monitor.writeToBuffer("Experiment", "populateDictionaryResults", "end");
  }

  /**
   * Populates the arrays of points for the different mangler vs relevance averages vs trial parameters; ignores groups
   * of manglers
   */
  private void populateManglerResults() {
    Global.monitor.writeToBuffer("Experiment", "populateManglerResults", "start");
    for (Entry<String, Double[]> mangler_result : this.manglerResultCounts.entrySet()) {
      double percent = 0;
      if (!mangler_result.getKey().equalsIgnoreCase("null")
          && mangler_result.getKey().indexOf('-') == mangler_result.getKey().lastIndexOf('-')) {
        percent = Double.valueOf(mangler_result.getKey().substring(mangler_result.getKey().indexOf('-') + 1));
      }
      if (this.manglerPlots.contains("deletetoken")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("deletetoken")) {
        this.deleteToken_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.deleteToken_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("deletechar")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("deletechar")) {
        this.deleteChar_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.deleteChar_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("deletewhitespace")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("deletewhitespace")) {
        this.deleteWhitespace_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.deleteWhitespace_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("changetochar")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("changetochar")) {
        this.changeToChar_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.changeToChar_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("changetoken")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("changetoken")) {
        this.changeToken_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.changeToken_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("splittoken")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("splittoken")) {
        this.splitToken_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.splitToken_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("changetopunc")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("changetopunc")) {
        this.changeToPunc_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.changeToPunc_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("deletesection")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("deletesection")) {
        this.deleteSection_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.deleteSection_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("deletehead")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("deletehead")) {
        this.deleteHead_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.deleteHead_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("deletetail")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("deletetail")) {
        this.deleteTail_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.deleteTail_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
      if (this.manglerPlots.contains("summaryreduce")
          && mangler_result.getKey().toLowerCase(Locale.ENGLISH).contains("summaryreduce")) {
        this.summaryReduce_precision.add(new XYDataItem(percent, mangler_result.getValue()[Global.PRECISION]
            / mangler_result.getValue()[Global.COUNT]));
        this.summaryReduce_recall.add(new XYDataItem(percent, mangler_result.getValue()[Global.RECALL]
            / mangler_result.getValue()[Global.COUNT]));
      }
    }
    Global.monitor.writeToBuffer("Experiment", "populateManglerResults", "end");
  }

  /**
   * Creates the different plots specified in the configuration file for this experiment
   * 
   * @param filename
   *          the string full path filename to save plots
   * @param colorScheme
   *          the <code>int</code> color scheme value (<code>Global.BLACK_WHITE</code> or <code>Global.COLOR</code>)
   */
  @SuppressWarnings("unchecked")
  private void createPlots(String filename, int colorScheme) {
    Global.monitor.writeToBuffer("Experiment", "createPlots", "start");
    for (String plot : this.plots) {
      if (plot.equalsIgnoreCase("precision_recall")) {
        Collections.sort(this.precision_recall);
        Map<String, List<XYDataItem>> map = Collections.singletonMap("precision-recall", this.precision_recall);
        savePlot(filename + "_precision-recall.pdf", this.dataset + ": Precision vs. Recall", "Precision", "Recall",
            colorScheme, false, createDataset(map));

      } else if (plot.equalsIgnoreCase("precision_fscore")) {
        Collections.sort(this.precision_fscore);
        Map<String, List<XYDataItem>> map = Collections.singletonMap("precision-fscore", this.precision_fscore);
        savePlot(filename + "_precision-fscore.pdf", this.dataset + ": Precision vs. F-Score", "Precision", "F-Score",
            colorScheme, false, createDataset(map));

      } else if (plot.equalsIgnoreCase("recall_fscore")) {
        Collections.sort(this.recall_fscore);
        Map<String, List<XYDataItem>> map = Collections.singletonMap("recall-fscore", this.recall_fscore);
        savePlot(filename + "_recall-fscore.pdf", this.dataset + ": Recall vs. F-Score", "Recall", "F-Score",
            colorScheme, false, createDataset(map));

      } else if (plot.equalsIgnoreCase("dictionary_precision")) {
        Collections.sort(this.dictionary_precision);
        Map<String, List<XYDataItem>> map = Collections.singletonMap("dictionary-precision", this.dictionary_precision);
        savePlot(filename + "_dictSize-precision.pdf", this.dataset + ": Dictionary Size vs. Average Precision",
            "Dictionary Size (uncompressed bits)", "Average Precision", colorScheme, false, createDataset(map));

      } else if (plot.equalsIgnoreCase("dictionary_recall")) {
        Collections.sort(this.dictionary_recall);
        Map<String, List<XYDataItem>> map = Collections.singletonMap("dictionary-recall", this.dictionary_recall);
        savePlot(filename + "_dictSize-recall.pdf", this.dataset + ": Dictionary Size vs. Average Recall",
            "Dictionary Size (uncompressed bits)", "Average Recall", colorScheme, false, createDataset(map));

      } else if (plot.equalsIgnoreCase("dictionary_fscore")) {
        Collections.sort(this.dictionary_fscore);
        Map<String, List<XYDataItem>> map = Collections.singletonMap("dictionary-fscore", this.dictionary_fscore);
        savePlot(filename + "_dictSize-fscore.pdf", this.dataset + ": Dictionary Size vs. Average F-Score",
            "Dictionary Size (uncompressed bits)", "Average F-Score", colorScheme, false, createDataset(map));

      } else if (plot.equalsIgnoreCase("dictionary_precision_recall")) {
        Collections.sort(this.dictionary_precision);
        Collections.sort(this.dictionary_recall);
        Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
        map.put("Average Precision", this.dictionary_precision);
        map.put("Average Recall", this.dictionary_recall);

        savePlot(filename + "_dictSize-precision-recall.pdf", this.dataset
            + ": Dictionary Size vs. Average Precision and Average Recall", "Dictionary Size (uncompressed bits)",
            "Average Relevance Scores", colorScheme, false, createDataset(map));

      } else if (plot.equalsIgnoreCase("mangler")) {
        if (this.manglerPlots.contains("deletetoken")) {
          Collections.sort(this.deleteToken_precision);
          Collections.sort(this.deleteToken_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.deleteToken_precision);
          map.put("Average Recall", this.deleteToken_recall);

          savePlot(filename + "_manglers_deleteToken.pdf", this.dataset
              + ": DeleteToken vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("deletechar")) {
          Collections.sort(this.deleteChar_precision);
          Collections.sort(this.deleteChar_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.deleteChar_precision);
          map.put("Average Recall", this.deleteChar_recall);

          savePlot(filename + "_manglers_deleteChar.pdf", this.dataset
              + ": DeleteChar vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("deletewhitespace")) {
          Collections.sort(this.deleteWhitespace_precision);
          Collections.sort(this.deleteWhitespace_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.deleteWhitespace_precision);
          map.put("Average Recall", this.deleteWhitespace_recall);

          savePlot(filename + "_manglers_deleteWhitespace.pdf", this.dataset
              + ": DeleteWhitespace vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("changetochar")) {
          Collections.sort(this.changeToChar_precision);
          Collections.sort(this.changeToChar_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.changeToChar_precision);
          map.put("Average Recall", this.changeToChar_recall);

          savePlot(filename + "_manglers_changeToChar.pdf", this.dataset
              + ": ChangeToChar vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("changetoken")) {
          Collections.sort(this.changeToken_precision);
          Collections.sort(this.changeToken_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.changeToken_precision);
          map.put("Average Recall", this.changeToken_recall);

          savePlot(filename + "_manglers_changeToken.pdf", this.dataset
              + ": ChangeToken vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("splittoken")) {
          Collections.sort(this.splitToken_precision);
          Collections.sort(this.splitToken_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.splitToken_precision);
          map.put("Average Recall", this.splitToken_recall);

          savePlot(filename + "_manglers_splitToken.pdf", this.dataset
              + ": SplitToken vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("changetopunc")) {
          Collections.sort(this.changeToPunc_precision);
          Collections.sort(this.changeToPunc_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.changeToPunc_precision);
          map.put("Average Recall", this.changeToPunc_recall);

          savePlot(filename + "_manglers_changeToPunc.pdf", this.dataset
              + ": ChangeToPunc vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("deletesection")) {
          Collections.sort(this.deleteSection_precision);
          Collections.sort(this.deleteSection_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.deleteSection_precision);
          map.put("Average Recall", this.deleteSection_recall);

          savePlot(filename + "_manglers_deleteSection.pdf", this.dataset
              + ": DeleteSection vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("deletehead")) {
          Collections.sort(this.deleteHead_precision);
          Collections.sort(this.deleteHead_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.deleteHead_precision);
          map.put("Average Recall", this.deleteHead_recall);

          savePlot(filename + "_manglers_deleteHead.pdf", this.dataset
              + ": DeleteHead vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("deletetail")) {
          Collections.sort(this.deleteTail_precision);
          Collections.sort(this.deleteTail_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.deleteTail_precision);
          map.put("Average Recall", this.deleteTail_recall);

          savePlot(filename + "_manglers_deleteTail.pdf", this.dataset
              + ": DeleteTail vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
        if (this.manglerPlots.contains("summaryreduce")) {
          Collections.sort(this.summaryReduce_precision);
          Collections.sort(this.summaryReduce_recall);
          Map<String, List<XYDataItem>> map = new HashMap<String, List<XYDataItem>>();
          map.put("Average Precision", this.summaryReduce_precision);
          map.put("Average Recall", this.summaryReduce_recall);

          savePlot(filename + "_manglers_summaryReduce.pdf", this.dataset
              + ": SummaryReduce vs. Average Precision and Average Recall", "Percent of File Deleted",
              "Average Relevance Scores", colorScheme, true, createDataset(map));
        }
      }
    }
    Global.monitor.writeToBuffer("Experiment", "createPlots", "end");
  } // createPlots

  /**
   * Creates a dataset containing x and y coordinate data
   * 
   * @param results
   *          the map containing the series labels and list of x and y coordinates to plot
   * 
   * @return the XYDataSet containing the series of x and y coordinate data
   */
  private static XYDataset createDataset(Map<String, List<XYDataItem>> results) {
    XYSeriesCollection dataset = new XYSeriesCollection();
    for (Entry<String, List<XYDataItem>> result : results.entrySet()) {
      XYSeries series = new XYSeries(result.getKey());
      for (XYDataItem point : result.getValue()) {
        series.add(point);
      }
      dataset.addSeries(series);
    }
    return dataset;
  }

  /**
   * Writes a scatter plot to a PDF, PNG, or JPEG file
   * 
   * @param outFilename
   *          the string name or path to store the resulting scatter plot; the extension of the filename is used to
   *          determine the type of file to write, PDF is the default
   * @param title
   *          the string title of the scatter plot
   * @param domainAxis
   *          the string label for the domain axis
   * @param rangeAxis
   *          the string label for the range axis
   * @param colorScheme
   *          the <code>int</code> color scheme value (<code>Global.BLACK_WHITE</code> or <code>Global.COLOR</code>)
   * @param lines
   *          the <code>boolean</code> value whether to connect related points with a line
   * @param dataset
   *          the xyDataset containing the points to include in this scatter plot
   */
  private static void savePlot(String outFilename, String title, String domainAxis, String rangeAxis, int colorScheme,
      boolean lines, XYDataset dataset) {

    XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer(lines, true);
    for (int i = 0; i < dataset.getSeriesCount(); i++) {
      if (colorScheme == BLACK_WHITE) {
        renderer.setSeriesPaint(i, GRAY_SCALE[i % GRAY_SCALE.length]);
      } else {
        renderer.setSeriesPaint(i, DARK_COLORS[i % DARK_COLORS.length]);
      }
      renderer.setSeriesShape(i, SHAPES[i % SHAPES.length]);
    }

    XYPlot plot = new XYPlot(dataset, new NumberAxis(domainAxis), new NumberAxis(rangeAxis), renderer);
    plot.setOrientation(PlotOrientation.VERTICAL);
    if (UIManager.get("Panel.background") == null) {
      UIManager.put("Panel.background", Color.WHITE);
    }
    JFreeChart chart = new JFreeChart(title, JFreeChart.DEFAULT_TITLE_FONT, plot, (dataset.getSeriesCount() > 1));

    int width = 640, height = 480;
    String outputFile = outFilename;
    if (outputFile.toLowerCase(Locale.ENGLISH).endsWith(".png")) {
      try {
        ChartUtilities.saveChartAsPNG(new File(outputFile), chart, width, height);
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    } else if (outputFile.toLowerCase(Locale.ENGLISH).endsWith(".jpeg")) {
      try {
        ChartUtilities.saveChartAsJPEG(new File(outputFile), chart, width, height);
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    } else {
      if (!outputFile.toLowerCase(Locale.ENGLISH).endsWith(".pdf")) {
        Global.monitor.printErrorMessage("Experiment", "Unrecognized file extension; creating PDF output file");
        Global.monitor.writeToBuffer("Experiment", "savePlot", "Unrecognized file extension; creating PDF output file");
        outputFile += ".pdf";
      }

      // Code help from: http://jtoee.blogspot.com/2008/04/writing-pdf-charts-with-itext-in-less.html
      com.lowagie.text.Document document = new com.lowagie.text.Document(new Rectangle(width, height));
      try {
        // Overwrites a file if one with the same name already exists
        PdfWriter pdfWriter = PdfWriter.getInstance(document, new FileOutputStream(outputFile));
        document.open();
        PdfContentByte pdfContentbyte = pdfWriter.getDirectContent();
        PdfTemplate pdfTemplate = pdfContentbyte.createTemplate(width, height);
        Graphics2D graphics = pdfTemplate.createGraphics(width, height, new DefaultFontMapper());
        Rectangle2D rectangle = new Rectangle2D.Double(0, 0, width, height);
        chart.draw(graphics, rectangle);
        graphics.dispose();
        pdfContentbyte.addTemplate(pdfTemplate, 0, 0);
      } catch (FileNotFoundException e) {
        ExceptionHandler.handleExceptionExit(e);
      } catch (DocumentException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
      document.close();
    }
  }

  /**
   * Runs experiments using multiple runs with different settings and generates results
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the experiment configuration file
   */
  public static void main(String args[]) {
    String config_file = null;
    try {
      config_file = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java Experiment <config_file>");
      System.err.println("Runs experiments using multiple runs with different settings and generates results");
      System.exit(-1);
    }

    try {
      Experiment exp = new Experiment(config_file);
      exp.showSettings();
      exp.runExperiment();
      Global.monitor.closeBuffer();
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
