package edu.georgetown.gucs.matcher;

import java.util.Arrays;

import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Determines if two fingerprints' documents are the same; uses a {@linkplain BitVectorFingerprinter}
 * 
 * @author Clay Shields
 */
public class ExactFingerprintMatcher extends FingerprintMatcher {

  /**
   * Constructor that sets the minimum score to use for matching two fingerprints to zero
   **/
  public ExactFingerprintMatcher() {
    super();
  }

  /**
   * Determines that the two fingerprints are matching if their byte arrays are equal
   */
  @Override
  public boolean match(byte[] fp1, byte[] fp2) {
    return Arrays.equals(fp1, fp2);
  }

  /**
   * Determines a similarity score for these two fingerprints
   * 
   * @return the similarity score between these two fingerprints; returns 0 if the fingerprints are not an exact match,
   *         returns 99 they are
   */
  @Override
  public int getScore(byte[] fp1, byte[] fp2) {
    if (match(fp1, fp2))
      return 99;
    return 0;
  }

  /**
   * Determines if two fingerprints' documents are the same; uses a bitVectorFingerprinter
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename containing the dictionary to use for fingerprinting<br>
   *          <code>args[1]</code> the filename of the document to fingerprint<br>
   *          <code>args[2]</code> the filename of the document to fingerprint
   */
  public static void main(String[] args) {
    String dictionary = null;
    String file1 = null;
    String file2 = null;

    try {
      dictionary = args[0];
      file1 = args[1];
      file2 = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java ExactFingerprintMatcher <dictionary> <fingeprint file> <fingeprint file>");
      System.exit(-1);
    }

    // Create two fingerprinters
    BitVectorFingerprinter fpr = new BitVectorFingerprinter(dictionary);
    ExactFingerprintMatcher efpm = new ExactFingerprintMatcher();

    for (Pair<String, String> fp1 : fpr.computeFingerprintXML(file1)) {
      for (Pair<String, String> fp2 : fpr.computeFingerprintXML(file2)) {
        String start_end1 = "";
        if (fp1.getB().length() > 0) {
          String[] start_endBytes = fp1.getB().split("\\s");
          start_end1 = " (start_byte: " + start_endBytes[0] + ", end_byte: " + start_endBytes[1] + ")";
        }
        System.out.println("File 1" + start_end1 + ": " + file1);

        String start_end2 = "";
        if (fp2.getB().length() > 0) {
          String[] start_endBytes = fp2.getB().split("\\s");
          start_end2 = " (start_byte: " + start_endBytes[0] + ", end_byte: " + start_endBytes[1] + ")";
        } 
        System.out.println("File 2" + start_end2 + ": " + file2);

        if (efpm.match(fp1.getA(), fp2.getA())) {
          System.out.println("MATCH");
        } else {
          System.out.println("NOT A MATCH");
        }
        System.out.println();
      }
    }
    Global.monitor.closeBuffer();
  }
}
