package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Dictionary;

import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.ar.ArabicAnalyzer;
import org.apache.lucene.analysis.tokenattributes.OffsetAttribute;
import org.apache.lucene.analysis.tokenattributes.TermAttribute;
import org.apache.lucene.util.Version;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Splits contents of an Arabic text file into tokens using Apache Lucene's {@linkplain ArabicAnalyzer}. Tokens are used
 * for {@linkplain Dictionary} creation.
 * 
 * @author Lindsay Neubauer
 */
public class ArabicFileTokenizer extends FileTokenizer {

  private static final long serialVersionUID = 3960239399901478440L;

  /**
   * Splits the document into tokens. Assumes the file is UTF-8 encoded and Version.LUCENE_30 for the ArabicAnalyzer.
   * 
   * @param filename
   *          the string filename of the Arabic language document to split into tokens
   * 
   * @see org.apache.lucene.analysis.ar.ArabicAnalyzer
   */
  @Override
  public void tokenize(String filename) {

    try {
      BufferedReader buffer_reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));
      ArabicAnalyzer arabicAnalyzer = new ArabicAnalyzer(Version.LUCENE_30);
      TokenStream tokenStream = arabicAnalyzer.tokenStream(null, buffer_reader);
      TermAttribute termAttribute = tokenStream.addAttribute(TermAttribute.class);
      OffsetAttribute offsetAttribute = tokenStream.addAttribute(OffsetAttribute.class);
      tokenStream.reset();

      synchronized (this.tokenVector) {
        synchronized (this.positions) {
          this.tokenVector.clear();
          this.positions.clear();

          while (tokenStream.incrementToken()) {
            String token = termAttribute.term();
            this.tokenVector.add(token);
            this.positions.add(new Pair<Integer, Integer>(offsetAttribute.startOffset(), offsetAttribute.endOffset()));
          }
        }
      }
      tokenStream.end();
      tokenStream.close();
      buffer_reader.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Tokenizes an Arabic text file and prints the resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the Arabic language document to split into tokens
   */
  public static void main(String[] args) {
    String input_file = "";
    try {
      input_file = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java -jar ArabicTokenizer.jar <file name>");
      System.err.println("Tokenizes an Arabic text file and prints the resulting tokens to the the screen");
      System.exit(-1);
    }

    try {
      ArabicFileTokenizer arabicTokenizer = new ArabicFileTokenizer();
      arabicTokenizer.tokenize(input_file);
      printTokens(arabicTokenizer.iterator());
      Global.monitor.closeBuffer();
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
