package edu.georgetown.gucs.tokenizers;

import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import net.sf.classifier4J.summariser.SimpleSummariser;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * Manipulates files as they are read. Settings are specified in an XML configuration file for a ParameterSearch and can
 * include:
 * <ul>
 * <li>null: makes no changes</li>
 * <li>deleteToken: deletes a given percentage of the tokens</li>
 * <li>deleteChar: deletes a given percentage of single characters</li>
 * <li>deleteWhitespace: deletes a given percentage of white space</li>
 * <li>deleteSection: deletes a given percentage of contiguous tokens</li>
 * <li>deleteHead: deletes a given percentage of contiguous tokens from the beginning of the file</li>
 * <li>deleteTail: deletes a given percentage of contiguous tokens from the end of the file</li>
 * <li>splitToken: inserts white space randomly inside a given percentage of tokens</li>
 * <li>changeToChar: changes a given percentage of characters to a random ascii lower case letter or number</li>
 * <li>changeToken: changes a given percentage of tokens to a random token from the dictionary</li>
 * <li>changeToPunc: changes a given percentage of characters to punctuation</li>
 * <li>summaryReduce: uses Classifier4J to create a document summary of a given percentage of the document size</li>
 * </ul>
 * 
 * @author Clay Shields
 **/
public class FileManglerTokenizer extends Tokenizer {

  private static final long serialVersionUID = 5750194440419510989L;
  private float deleteToken;
  private float deleteSection;
  private float deleteHead;
  private float deleteTail;
  private float deleteChar;
  private float deleteWhitespace;
  private float changeToChar;
  private float changeToPunc;
  private float changeToken;
  private float splitToken;
  private float summaryReduce;
  private Random random;
  private List<String> dictionary;
  private boolean nullProcess;
  private String tokensAsString;

  /**
   * Constructor that initializes the random number generator and clears the mangler settings
   */
  public FileManglerTokenizer() {
    this.random = new Random();
    this.nullProcess = true;
    this.tokensAsString = null;
    clearManglerSettings();
  }

  /**
   * Constructor that sets the mangler settings
   * 
   * @param manglers
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   */
  public FileManglerTokenizer(String manglers) {
    this.random = new Random();
    this.tokensAsString = null;
    this.nullProcess = false;
    parseSettings(manglers);
  }

  /**
   * Turns off (sets to zero) all the mangler settings for this tokenizer
   */
  private void clearManglerSettings() {
    this.deleteToken = 0;
    this.deleteSection = 0;
    this.deleteHead = 0;
    this.deleteTail = 0;
    this.deleteChar = 0;
    this.deleteWhitespace = 0;
    this.changeToChar = 0;
    this.changeToPunc = 0;
    this.changeToken = 0;
    this.splitToken = 0;
    this.summaryReduce = 0;
  }

  /**
   * Enables the mangler with the given settings in this tokenizer
   * 
   * @param manglers
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   */
  public void enableMangler(String manglers) {
    this.nullProcess = false;
    parseSettings(manglers);
  }

  /**
   * Enables the mangler with the given settings in this tokenizer
   * 
   * @param manglers
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   * @param tokens
   *          the list of string tokens to put through this mangler
   */
  public void enableMangler(String manglers, List<String> tokens) {
    this.dictionary = tokens;
    this.nullProcess = false;
    parseSettings(manglers);
  }

  /**
   * Disables the mangler in this tokenizer
   */
  public void disableMangler() {
    this.nullProcess = true;
    clearManglerSettings();
  }

  /**
   * Sets the random number generator to use with the manglers that are set in this tokenizer
   * 
   * @param random
   *          the random number generator to use with this tokenizer
   */
  public void setRNG(Random random) {
    this.random = random;
  }

  /**
   * Sets each mangler setting present in the given list; throws an error if the same mangler is present more than once
   * 
   * @param manglers
   *          the string of mangler settings; expects non-null mangler strings in the form:
   *          "manglerName-manglerPercent", where the manglerPercent is between 1 and 99 inclusive
   */
  private void parseSettings(String manglers) {
    clearManglerSettings();
    if (manglers.contains("null")) {
      disableMangler();
    } else {
      String[] manglerArray = manglers.split("\\s");
      for (String manglerInput : manglerArray) {
        String[] manglerSplit = manglerInput.split("-");
        String mangler = manglerSplit[0];
        float percent = Float.valueOf(manglerSplit[1]);
        if (percent < 1 || percent > 99) {
          ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: percent must be between 1-99"));
        }

        if (mangler.equalsIgnoreCase("deleteToken")) {
          if (this.deleteToken > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple deleteToken calls"));
          this.deleteToken = percent;
        } else if (mangler.equalsIgnoreCase("deleteSection")) {
          if (this.deleteSection > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple deleteSection calls"));
          this.deleteSection = percent;
        } else if (mangler.equalsIgnoreCase("deleteHead")) {
          if (this.deleteHead > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple deleteHead calls"));
          this.deleteHead = percent;
        } else if (mangler.equalsIgnoreCase("deleteTail")) {
          if (this.deleteTail > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple deleteTail calls"));
          this.deleteTail = percent;
        } else if (mangler.equalsIgnoreCase("deleteChar")) {
          if (this.deleteChar > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple deleteChar calls"));
          this.deleteChar = percent;
        } else if (mangler.equalsIgnoreCase("deleteWhitespace")) {
          if (this.deleteWhitespace > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple deleteWhitespace calls"));
          this.deleteWhitespace = percent;
        } else if (mangler.equalsIgnoreCase("changeToChar")) {
          if (this.changeToChar > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple changeToChar calls"));
          this.changeToChar = percent;
        } else if (mangler.equalsIgnoreCase("changeToken")) {
          if (this.changeToken > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple changeToken calls"));
          this.changeToken = percent;
        } else if (mangler.equalsIgnoreCase("splitToken")) {
          if (this.splitToken > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple splitToken calls"));
          this.splitToken = percent;
        } else if (mangler.equalsIgnoreCase("changeToPunc")) {
          if (this.changeToPunc > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple changeToPunc calls"));
          this.changeToPunc = percent;
        } else if (mangler.equalsIgnoreCase("summaryReduce")) {
          if (this.summaryReduce > 0)
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: multiple summaryReduce calls"));
          this.summaryReduce = percent;
        }
      }
    }
  } // end parseSettings

  /**
   * Prints mangler settings to standard output
   */
  public void showSettings() {
    System.out.println("nullProcess: " + this.nullProcess);
    System.out.println("deleteToken: " + this.deleteToken);
    System.out.println("deleteChar: " + this.deleteChar);
    System.out.println("deleteWhitespace: " + this.deleteWhitespace);
    System.out.println("changeToChar: " + this.changeToChar);
    System.out.println("changeToken: " + this.changeToken);
    System.out.println("splitToken: " + this.splitToken);
    System.out.println("changeToPunc: " + this.changeToPunc);
    System.out.println("deleteSection: " + this.deleteSection);
    System.out.println("deleteHead: " + this.deleteHead);
    System.out.println("deleteTail: " + this.deleteTail);
    System.out.println("summaryReduce: " + this.summaryReduce);
  }

  /**
   * Alters or eliminates certain tokens based on the given mangler settings
   */
  @Override
  public void tokenize(Iterator<String> iterator) {
    synchronized (this.tokenVector) {
      this.tokenVector.clear();
      while (iterator.hasNext()) {
        String token = iterator.next().trim();
        if (token.length() > 0)
          this.tokenVector.add(token);
      }

      // Apply mangler settings
      if (!this.nullProcess) {

        // Deletes a given percentage of the tokens
        if (this.deleteToken > 0) {
          int deletionSize = (int) ((this.deleteToken / 100) * this.tokenVector.size());
          for (int i = 0; i < deletionSize; i++) {
            this.tokenVector.remove(this.random.nextInt(this.tokenVector.size()));
          }
        }

        // Deletes a given percentage of contiguous tokens
        if (this.deleteSection > 0) {
          int deletionSize = (int) ((this.deleteSection / 100) * this.tokenVector.size());
          int maxStart = this.tokenVector.size() - deletionSize;
          if (maxStart < 0) {
            String exception = "FileManglerTokenizer: DeleteSection is larger than token vector" + ": vector size= "
                + this.tokenVector.size() + " setting in pct " + this.deleteSection + " num to delete " + deletionSize;
            ExceptionHandler.handleExceptionThrow(new Exception(exception));
          }
          int randomStart = (maxStart == 0) ? 0 : this.random.nextInt(maxStart);
          this.tokenVector.subList(randomStart, randomStart + deletionSize).clear();
        }

        // Deletes a given percentage of contiguous tokens from the beginning of
        // the file
        if (this.deleteHead > 0) {
          int deletionSize = (int) ((this.deleteHead / 100) * this.tokenVector.size());
          this.tokenVector.subList(0, deletionSize).clear();
        }

        // Deletes a given percentage of contiguous tokens from the end of the
        // file
        if (this.deleteTail > 0) {
          int deletionSize = (int) ((this.deleteTail / 100) * this.tokenVector.size());
          this.tokenVector.subList(this.tokenVector.size() - deletionSize, this.tokenVector.size()).clear();
        }

        // Deletes a given percentage of single characters
        if (this.deleteChar > 0) {
          int deletionSize = (int) ((this.deleteChar / 100) * this.tokenVector.size());
          for (int i = 0; i < deletionSize; i++) {
            StringBuffer token = new StringBuffer(this.tokenVector.remove(this.random.nextInt(this.tokenVector.size())));
            if (token.length() > 0) {
              token.deleteCharAt(this.random.nextInt(token.length()));
              this.tokenVector.add(token.toString());
            }
          }
        }

        // Deletes a given percentage of white space
        if (this.deleteWhitespace > 0) {
          int deletionSize = (int) ((this.deleteWhitespace / 100) * this.tokenVector.size());
          for (int i = 0; i < deletionSize; i++) {
            int position = this.random.nextInt(this.tokenVector.size() - 1);

            // For each gap between tokens, there is a deleteWhitespace percent
            // chance of it being deleted
            String newToken = this.tokenVector.get(position) + this.tokenVector.get(position + 1);
            this.tokenVector.remove(position + 1);
            this.tokenVector.set(position, newToken);
          }
        }

        // Changes a given percentage of characters to a random ASCII lower case
        // letter or number
        if (this.changeToChar > 0) {
          int deletionSize = (int) ((this.changeToChar / 100) * this.tokenVector.size());
          for (int i = 0; i < deletionSize; i++) {
            StringBuffer token = new StringBuffer(this.tokenVector.remove(this.random.nextInt(this.tokenVector.size())));
            token.setCharAt(this.random.nextInt(token.length()), (char) (this.random.nextInt(26) + 97));
            this.tokenVector.add(token.toString());
          }
        }

        // Changes a given percentage of tokens to a random token from the
        // dictionary
        if (this.changeToken > 0) {

          if (this.dictionary == null || this.dictionary.size() == 0) {
            ExceptionHandler.handleExceptionThrow(new Exception("FileManglerTokenizer: null or empty dictionary"));
          }

          int deletionSize = (int) ((this.changeToken / 100) * this.tokenVector.size());
          for (int i = 0; i < deletionSize; i++) {
            this.tokenVector.set(this.random.nextInt(this.tokenVector.size()),
                this.dictionary.get(this.random.nextInt(this.dictionary.size())));
          }
        }

        // Changes a given percentage of characters to punctuation
        if (this.changeToPunc > 0) {
          // create a vector of punctuation symbols
          Vector<Character> punc = new Vector<Character>();
          for (char i = 33; i <= 47; i++) {
            punc.add(i);
          }
          for (char i = 58; i <= 64; i++) {
            punc.add(i);
          }
          for (char i = 91; i <= 96; i++) {
            punc.add(i);
          }
          for (char i = 123; i <= 126; i++) {
            punc.add(i);
          }

          int deletionSize = (int) ((this.changeToPunc / 100) * this.tokenVector.size());
          for (int i = 0; i < deletionSize; i++) {
            StringBuffer token = new StringBuffer(this.tokenVector.remove(this.random.nextInt(this.tokenVector.size())));
            token.setCharAt(this.random.nextInt(token.length()), punc.get(this.random.nextInt(punc.size())));
            this.tokenVector.add(token.toString());
          }
        }

        // Inserts white space randomly inside a given percentage of tokens
        if (this.splitToken > 0) {
          int deletionSize = (int) ((this.splitToken / 100) * this.tokenVector.size());
          Vector<String> tempVector = new Vector<String>();
          for (int i = 0; i < deletionSize; i++) {
            String token = this.tokenVector.remove(this.random.nextInt(this.tokenVector.size()));
            if (token.length() > 1) {
              int splitPoint = this.random.nextInt(token.length());
              // Don't want to split the same token twice
              tempVector.add(token.substring(0, splitPoint));
              tempVector.add(token.substring(splitPoint));
            }
          }
          // Add all split tokens back into tokenVector
          this.tokenVector.addAll(tempVector);
        }

        // Uses Classifier4J to create a document summary of a given percentage of
        // the document size
        if (this.summaryReduce > 0) {
          // Turn the vector of tokens into a long string and find out how many
          // sentences there are
          int sentenceCount = tokensToSentences();

          // Figure out how many sentences to keep
          int sentenceToKeep = (int) (sentenceCount * ((100 - this.summaryReduce) / 100));

          SimpleSummariser summariser = new SimpleSummariser();
          String summary = summariser.summarise(this.tokensAsString, sentenceToKeep);

          // Split the result into a new tokenVector and replace the one we were
          // using
          Vector<String> newTokenVector = new Vector<String>();
          for (String result : summary.split("\\s")) {
            if (result.trim().length() > 0) {
              newTokenVector.add(result.trim());
            }
          }
          this.tokenVector.clear();
          this.tokenVector.addAll(newTokenVector);
        }
      }
    }
  }

  /**
   * Processes tokens into sentences for use with the summariser; Does special processing when first line contains
   * "Message-ID" (looks like an email); Puts period at end of each header line to clearly mark it as the end of line
   * 
   * @return the <code>int</code> number of sentences
   */
  private int tokensToSentences() {
    // This is a hack: Probably should have rewritten the File Tokenizer to do
    // this
    boolean inHeaders = (this.tokenVector.get(0).contains("Message-ID:")) ? true : false;

    StringBuffer fullFile = new StringBuffer(this.tokenVector.get(0));
    int sentenceCount = 0;

    for (int i = 1; i < this.tokenVector.size(); i++) {
      fullFile.append(" " + this.tokenVector.get(i));

      if (inHeaders) {
        // peek ahead at the next token to see if it has a :
        if ((i < this.tokenVector.size() - 2) && (this.tokenVector.get(i + 1).contains(":"))) {
          // if so, add a . to the current line
          fullFile.append(".");
          sentenceCount++;
        }
        // If at the X-FileName line should be the end of the headers
        if ((this.tokenVector.get(i).contains("X-FileName:"))) {
          // A token that has a . in it is probably the file name
          if (this.tokenVector.get(i).contains(".")) {
            inHeaders = false;
            fullFile.append(".");
            sentenceCount++;
          }
        }
      }

      // Check if we should count a new sentence
      // When we see a token that ends with a . ! or ?
      int length = this.tokenVector.get(i).length();

      // If it is a blank line, skip this check
      if (length != 0) {
        // char last = tokenVector.get(i).charAt(length - 1);
        if ((this.tokenVector.get(i).charAt(length - 1) == '.') || (this.tokenVector.get(i).charAt(length - 1) == '!')
            || (this.tokenVector.get(i).charAt(length - 1) == '?')) {
          sentenceCount++;
        }
      }
    }// end for i

    this.tokensAsString = fullFile.toString();
    return sentenceCount;
  }
}
