package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Dictionary;
import java.util.Locale;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Splits contents of a text file into tokens based on whitespace or by line. Used for {@linkplain Dictionary} creation.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 **/
public class FileTokenizer extends Tokenizer {

  private static final long serialVersionUID = 7140999692348553037L;
  /**
   * The string specifying to split tokens based on whitespace ("tokens") or by line ("lines")
   */
  protected String mode;

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public FileTokenizer() {
    this.mode = "tokens";
  }

  /**
   * Constructor that sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the
   * file is split based on whitespace.
   * 
   * @param mode
   *          the string specifying to split tokens based on whitespace or by line
   */
  public FileTokenizer(String mode) {
    setMode(mode);
  }

  /**
   * Sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the file is split
   * based on whitespace.
   * 
   * @param mode
   *          the string specifying to split tokens based on whitespace or by line
   */
  public void setMode(String mode) {
    if (mode.trim().equalsIgnoreCase("tokens") || mode.trim().equalsIgnoreCase("lines")) {
      this.mode = mode.toLowerCase(Locale.ENGLISH).trim();
    } else {
      ExceptionHandler.handleExceptionThrow(new Exception("FileTokenizer: Invalid mode (" + mode + ")"));
    }
  }

  /**
   * Provides the token creation mode
   * 
   * @return the string specifying to split tokens based on whitespace or by line
   */
  public String getMode() {
    return this.mode;
  }

  /**
   * Splits the document into tokens.
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   */
  @Override
  public void tokenize(String filename) {
    readFile(filename);
  }

  /**
   * Splits a text document into tokens.
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   */
  protected void readFile(String filename) {
    BufferedReader buffer_reader = null;
    try {
      // The encoding here needs to match the encoding of the file being read
      // This is assuming the file is UTF-8 encoded (otherwise, it will spit out garbage)
      buffer_reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));

      StringBuffer token = new StringBuffer("");
      int character;
      int byte_start = 0;
      int byte_end = 0;

      synchronized (this.tokenVector) {
        synchronized (this.positions) {
          this.tokenVector.clear();
          this.positions.clear();
          while ((character = buffer_reader.read()) != -1) {

            if (this.mode.equalsIgnoreCase("lines")) {
              if ('\n' == character || '\r' == character || '\f' == character) {
                if (token.toString().trim().length() > 0) {
                  this.tokenVector.add(token.toString().trim());
                  this.positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
                  byte_start = byte_end;

                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            } else {
              if (Character.isWhitespace(character)) {
                if (token.toString().trim().length() > 0) {
                  this.tokenVector.add(token.toString().trim());
                  this.positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
                  byte_start = byte_end;
                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            }
          }
          buffer_reader.close();
          if (token.toString().trim().length() > 0) {
            this.tokenVector.add(token.toString().trim());
            this.positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
          }
        }
      }
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } finally {
      try {
        if (buffer_reader != null) {
          buffer_reader.close();
        }
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
  }

  /**
   * Prints to the screen a portion of the given text document
   * 
   * @param filename
   *          the string filename of the document to split into tokens
   * @param start_index
   *          the number of characters to skip in file before printing, inclusive
   * @param end_index
   *          the number of characters to print from the file, exclusive
   */
  public void checkIndexing(String filename, int start_index, int end_index) {
    try {
      BufferedReader buffer_reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "UTF-8"));
      long byte_position = buffer_reader.skip(start_index);
      StringBuffer str = new StringBuffer("");
      while (byte_position < end_index) {
        int character = buffer_reader.read();
        str.append((char) character);
        byte_position += Character.charCount(character);
      }
      System.out.println(str.toString());
      buffer_reader.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Tokenizes a text file and prints the resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the English language document to split into tokens
   */
  public static void main(String args[]) {
    String filename = "";
    try {
      filename = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java -jar FileTokenizer.jar <file name>");
      System.err.println("Tokenizes a text file and prints the resulting tokens to the the screen");
      System.exit(-1);
    }

    try {
      FileTokenizer fileTokenizer = new FileTokenizer();
      fileTokenizer.tokenize(filename);
      printTokens(fileTokenizer.iterator());
      Global.monitor.closeBuffer();
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }

  }
}
