package edu.georgetown.gucs.tokenizers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Partitions a document into multiple pieces before applying a list of (optional) {@linkplain Tokenizer} objects.
 * 
 * @author Lindsay Neubauer
 * 
 */
public class Splitter extends FileTokenizer {
  private static final long serialVersionUID = 7671055598312124506L;
  private TokenizerList fileTokenizer;
  private TokenizerList tokenizers;
  private List<String> splitterNames;
  private Element root;

  /**
   * Constructor that sets the tokenizers and splitters to use on this document. The first tokenizer in the
   * tokenizerList must supports reading from a file.
   * 
   * @param tokenizerList
   *          the string list of the tokenizers for this splitter
   * @param splittersElement
   *          the XML element containing the splitters
   */
  @SuppressWarnings("unchecked")
  public Splitter(List<String> tokenizerList, Element splittersElement) {
    this.splitterNames = new Vector<String>();

    this.root = new Element("splitter");
    this.root.setAttribute("creation", String.valueOf(new Date()));

    this.fileTokenizer = new TokenizerList(tokenizerList.subList(0, 1));
    this.tokenizers = new TokenizerList(tokenizerList.subList(1, tokenizerList.size()));

    for (Element splitter : (List<Element>) splittersElement.getChildren("splitter")) {
      this.splitterNames.add(splitter.getAttributeValue("name") + " " + splitter.getAttributeValue("value"));
    }
  }

  /**
   * Constructor that sets the tokenizers and splitters from a configuration file to use on this document. The first
   * tokenizer in the configuration file must supports reading from a file.
   * 
   * @param config_file
   *          the string XML filename that specifies these tokenizers
   */
  @SuppressWarnings("unchecked")
  public Splitter(String config_file) {
    this.splitterNames = new Vector<String>();

    this.root = new Element("splitter");
    this.root.setAttribute("creation", String.valueOf(new Date()));

    try {
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);
      Element splittersElement = (doc.getRootElement().getName().equalsIgnoreCase("splitters")) ? doc.getRootElement()
          : doc.getRootElement().getChild("splitters");

      for (Element splitter : (List<Element>) splittersElement.getChildren("splitter")) {
        this.splitterNames.add(splitter.getAttributeValue("name") + " " + splitter.getAttributeValue("value"));
      }

      List<String> tokenizerNames = new Vector<String>();
      for (Element tokenizerElement : (List<Element>) splittersElement.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        Element tokeElement = new Element("tokenizer");
        tokeElement.setAttribute("name", tokenizer);
        if (this.fileTokenizer == null && !tokenizer.contains("FileTokenizer")) {
          ExceptionHandler.handleExceptionThrow(new Exception("Experiment: First tokenizer (" + tokenizer
              + ") must read from a file."));
        }

        if (tokenizerElement.getAttributeValue("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
          tokeElement.setAttribute("parameter", tokenizerElement.getAttributeValue("parameter"));
        }
        this.root.addContent(tokeElement);
        tokenizerNames.add(tokenizer.trim());

        if (tokenizer.contains("FileTokenizer") && this.fileTokenizer == null) {
          this.fileTokenizer = new TokenizerList(tokenizerNames);
          tokenizerNames.clear();
        } else if (tokenizer.contains("FileTokenizer") && this.fileTokenizer != null) {
          ExceptionHandler.handleExceptionThrow(new Exception("Splitter: Cannot use more than one FileTokenizer"));
        }
      }
      this.tokenizers = new TokenizerList(tokenizerNames);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Provides the list of splitters
   * 
   * @return the list of splitters
   */
  public List<String> getSplitterNames() {
    return this.splitterNames;
  }

  /**
   * Provides the list of splitters
   * 
   * @return the list of splitters
   */
  public List<String> getTokenizersNames() {
    List<String> allTokenizers = this.fileTokenizer.getNames();
    allTokenizers.addAll(this.tokenizers.getNames());
    return allTokenizers;
  }

  @Override
  public void tokenize(String filename) {
    this.root.setAttribute("file", (new File(filename)).getAbsolutePath());
    this.fileTokenizer.tokenizeFile(filename);
    this.tokenVector = this.fileTokenizer.getTokenizers().get(0).tokenVector;
    this.positions = this.fileTokenizer.getTokenizers().get(0).positions;

    for (String splitter_name : this.splitterNames) {
      if (splitter_name.contains("proportions")) {
        proportion(Double.valueOf(splitter_name.substring(splitter_name.indexOf(' ') + 1)));
      } else if (splitter_name.contains("chunks")) {
        fixed(Integer.valueOf(splitter_name.substring(splitter_name.indexOf(' ') + 1)));
      } else if (splitter_name.contains("w-shingling")) {
        if (splitter_name.contains("unique")) {
          w_shingleUnique(Integer.valueOf(splitter_name.substring(splitter_name.indexOf(' ') + 1)));
        } else {
          w_shingle(Integer.valueOf(splitter_name.substring(splitter_name.indexOf(' ') + 1)));
        }
      } else {
        ExceptionHandler.handleExceptionThrow(new Exception("Splitter: Inproper inputs given to tokenize: "
            + splitter_name));
      }
    }
  }

  /**
   * Writes the results of splitting to an XML file
   * 
   * @param outputfile
   *          the string filename of the document to write the results of splitting to
   */
  public void writeDigest(String outputfile) {
    try {
      Document doc = new Document(this.root);
      FileOutputStream dictionaryOut = new FileOutputStream(outputfile);
      Format form = Format.getPrettyFormat();
      form.setEncoding("UTF-8");
      XMLOutputter out = new XMLOutputter(form);
      out.output(doc, dictionaryOut);
      dictionaryOut.close();
      System.out.println("Digest written to: " + outputfile);
      Global.monitor.writeToBuffer("Splitter", "writeDigest", "Digest written to: " + outputfile);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleExceptionExit(e);
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Splits the document into chunks based on the given proportion. Values between 0 and 1 split the document by a
   * percentage. Values >= 1 are truncated and split the document into that many parts.
   * 
   * @param proportion
   *          the <code>double</code> proportion to split the document into; values between 0 and 1 split the document
   *          by that percentage, values >= 1 are truncated and split the document into that many parts
   */
  private void proportion(double proportion) {

    Element splitter = new Element("proportion");
    splitter.setAttribute("value", String.valueOf(proportion));

    int iteration = 1;
    int start = 0;
    int end;
    Iterator<String> tokensIterator;
    Iterator<Pair<Integer, Integer>> positionsIterator;
    Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>> pair;
    while (start < this.tokenVector.size()) {
      if (proportion < 1) {
        // each part is that proportion of the document
        end = (int) (iteration * proportion * this.tokenVector.size());
      } else {
        // split document into that many parts
        end = (int) (this.tokenVector.size() / proportion * iteration);
      }

      if (end < this.tokenVector.size()) {
        pair = this.tokenizers.split_tokenize(this.tokenVector.subList(start, end).iterator(),
            this.positions.subList(start, end).iterator());
      } else {
        pair = this.tokenizers.split_tokenize(this.tokenVector.subList(start, this.tokenVector.size()).iterator(),
            this.positions.subList(start, this.tokenVector.size()).iterator());
      }
      tokensIterator = pair.getA();
      positionsIterator = pair.getB();

      // Add to digest
      Element split = new Element("split");
      if (positionsIterator.hasNext()) {
        Pair<Integer, Integer> position = positionsIterator.next();
        split.setAttribute("name", "proportion");
        split.setAttribute("iteration", String.valueOf(iteration));
        split.setAttribute("start_byte", String.valueOf(position.getA()));
        while (tokensIterator.hasNext()) {
          Element token = new Element("token");
          token.setAttribute("name", tokensIterator.next());
          token.setAttribute("start_byte", String.valueOf(position.getA()));
          token.setAttribute("end_byte", String.valueOf(position.getB()));
          split.addContent(token);
          if (positionsIterator.hasNext()) {
            position = positionsIterator.next();
          }
        }
        split.setAttribute("end_byte", String.valueOf(position.getB()));
        splitter.addContent(split);
      }
      start = end;
      iteration++;
    }
    this.root.addContent(splitter);
  }

  /**
   * Splits the document into chunks of tokens of the given size. One chunk (numTokens = tokenVector.size()-1) is
   * created if the given number of tokens is larger than the number of tokens in the document.
   * 
   * @param numTokens
   *          the <code>int</code> number of tokens to put in each portion of the document
   */
  private void fixed(int numTokens) {
    if (numTokens >= this.tokenVector.size()) {
      Global.monitor.printErrorMessage("Splitter", "Given chunk size (" + numTokens
          + ") exceeds the number of tokens in the document. Creating one chunk with size "
          + (this.tokenVector.size() - 1));
      Global.monitor.writeToBuffer("Splitter", "fixed", "Given chunk size (" + numTokens
          + ") exceeds the number of tokens in the document. Creating one chunk with size "
          + (this.tokenVector.size() - 1));
      numTokens = this.tokenVector.size() - 1;
    }

    Element splitter = new Element("chunks");
    splitter.setAttribute("size", String.valueOf(numTokens));

    int start = 0;
    Iterator<String> tokensIterator;
    Iterator<Pair<Integer, Integer>> positionsIterator;
    Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>> pair;
    for (int iteration = 1; start < this.tokenVector.size(); iteration++) {
      if (start + numTokens < this.tokenVector.size()) {
        pair = this.tokenizers.split_tokenize(this.tokenVector.subList(start, start + numTokens).iterator(),
            this.positions.subList(start, start + numTokens).iterator());
      } else {
        pair = this.tokenizers.split_tokenize(this.tokenVector.subList(start, this.tokenVector.size()).iterator(),
            this.positions.subList(start, this.tokenVector.size()).iterator());
      }
      tokensIterator = pair.getA();
      positionsIterator = pair.getB();

      // Add to digest
      Element split = new Element("split");
      Pair<Integer, Integer> position;
      if (positionsIterator.hasNext()) {
        position = positionsIterator.next();
        split.setAttribute("name", "chunks");
        split.setAttribute("iteration", String.valueOf(iteration));
        split.setAttribute("start_byte", String.valueOf(position.getA()));
        while (tokensIterator.hasNext()) {
          Element token = new Element("token");
          token.setAttribute("name", tokensIterator.next());
          token.setAttribute("start_byte", String.valueOf(position.getA()));
          token.setAttribute("end_byte", String.valueOf(position.getB()));
          split.addContent(token);
          if (positionsIterator.hasNext())
            position = positionsIterator.next();
        }
        split.setAttribute("end_byte", String.valueOf(position.getB()));
        splitter.addContent(split);
      }

      start += numTokens;
    }
    this.root.addContent(splitter);
  }

  /**
   * Splits the document into chunks using w-shingling. Each chunk is a contiguous subsequence of tokens in the document
   * of the given size. One chunk (nGrams = tokenVector.size()-1) is created if the given number of tokens is larger
   * than the number of tokens in the document.
   * 
   * @param nGrams
   *          the <code>int</code> number of tokens to put in each portion of the document
   */
  private void w_shingle(int nGrams) {
    if (nGrams >= this.tokenVector.size()) {
      Global.monitor.printErrorMessage("Splitter", "Given n-gram size (" + nGrams
          + ") exceeds the number of tokens in the document. Creating one chunk with n-gram = size = "
          + (this.tokenVector.size() - 1));
      Global.monitor.writeToBuffer("Splitter", "w_shingle", "Given n-gram size (" + nGrams
          + ") exceeds the number of tokens in the document. Creating one chunk with n-gram = size = "
          + (this.tokenVector.size() - 1));
      nGrams = this.tokenVector.size() - 1;
    }

    Element splitter = new Element("w-shingle");
    splitter.setAttribute("N-grams", String.valueOf(nGrams));

    int end = nGrams;
    Iterator<String> tokensIterator;
    Iterator<Pair<Integer, Integer>> positionsIterator;
    Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>> pair;
    for (int start = 0; end <= this.tokenVector.size(); start++, end++) {
      pair = this.tokenizers.split_tokenize(this.tokenVector.subList(start, end).iterator(),
          this.positions.subList(start, end).iterator());

      tokensIterator = pair.getA();
      positionsIterator = pair.getB();

      // Add to digest
      Element split = new Element("split");
      Pair<Integer, Integer> position;
      if (positionsIterator.hasNext()) {
        position = positionsIterator.next();
        split.setAttribute("name", "w-shingle");
        split.setAttribute("iteration", String.valueOf(start + 1));
        split.setAttribute("start_byte", String.valueOf(position.getA()));
        while (tokensIterator.hasNext()) {
          Element token = new Element("token");
          token.setAttribute("name", tokensIterator.next());
          token.setAttribute("start_byte", String.valueOf(position.getA()));
          token.setAttribute("end_byte", String.valueOf(position.getB()));
          split.addContent(token);
          if (positionsIterator.hasNext())
            position = positionsIterator.next();
        }
        split.setAttribute("end_byte", String.valueOf(position.getB()));
        splitter.addContent(split);
      }
    }
    this.root.addContent(splitter);
  }

  /**
   * Splits the document into chunks using unique w-shingling. Each chunk is a contiguous and unique subsequence of
   * tokens in the document of the given size. One chunk (nGrams = tokenVector.size()-1) is created if the given number
   * of tokens is larger than the number of tokens in the document.
   * <p>
   * The first occurrence of the subsequence of tokens is stored.
   * <p>
   * Unique w-shingling occurs before other tokenizers are applied so duplicate shingles are possible.
   * 
   * @param nGrams
   *          the <code>int</code> number of tokens to put in each portion of the document
   */
  private void w_shingleUnique(int nGrams) {
    if (nGrams >= this.tokenVector.size()) {
      Global.monitor.printErrorMessage("Splitter", "Given n-gram size (" + nGrams
          + ") exceeds the number of tokens in the document. Creating one chunk with n-gram = size = "
          + (this.tokenVector.size() - 1));
      Global.monitor.writeToBuffer("Splitter", "w_shingleUnique", "Given n-gram size (" + nGrams
          + ") exceeds the number of tokens in the document. Creating one chunk with n-gram = size = "
          + (this.tokenVector.size() - 1));
      nGrams = this.tokenVector.size() - 1;
    }

    Element splitter = new Element("w-shingle-unique");
    splitter.setAttribute("N-grams", String.valueOf(nGrams));

    int end = nGrams;
    Iterator<String> tokensIterator;
    Iterator<Pair<Integer, Integer>> positionsIterator;
    Pair<Iterator<String>, Iterator<Pair<Integer, Integer>>> pair;
    Set<List<String>> tokenSet = new LinkedHashSet<List<String>>();
    int tokenSet_before, tokenSet_after = 0;
    for (int start = 0; end <= this.tokenVector.size(); start++, end++) {
      pair = this.tokenizers.split_tokenize(this.tokenVector.subList(start, end).iterator(),
          this.positions.subList(start, end).iterator());
      tokensIterator = pair.getA();
      positionsIterator = pair.getB();

      // Add to digest
      Element split = new Element("split");
      Pair<Integer, Integer> position;
      List<String> updatedTokenVector = new Vector<String>();
      tokenSet_before = tokenSet.size();
      if (positionsIterator.hasNext()) {
        position = positionsIterator.next();
        split.setAttribute("name", "w-shingle-unique");
        split.setAttribute("iteration", String.valueOf(start + 1));
        split.setAttribute("start_byte", String.valueOf(position.getA()));
        while (tokensIterator.hasNext()) {
          Element token = new Element("token");
          String tokenName = tokensIterator.next();
          updatedTokenVector.add(tokenName);
          token.setAttribute("name", tokenName);
          token.setAttribute("start_byte", String.valueOf(position.getA()));
          token.setAttribute("end_byte", String.valueOf(position.getB()));
          split.addContent(token);
          if (positionsIterator.hasNext())
            position = positionsIterator.next();
        }
        split.setAttribute("end_byte", String.valueOf(position.getB()));
        tokenSet.add(updatedTokenVector);
        tokenSet_after = tokenSet.size();
        if (tokenSet_before < tokenSet_after) {
          splitter.addContent(split);
        }
      }
    }
    this.root.addContent(splitter);
  }

  /**
   * Tokenizes different chunks of a text file and stores them in an output XML file
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the document to split into tokens<br>
   *          <code>args[1]</code> the filename of the splitter configuration file<br>
   *          <code>args[2]</code> the filename to write the XML output<br>
   */
  public static void main(String args[]) {
    String input_file = "";
    String config_file = "";
    String output_file = "";
    try {
      input_file = args[0];
      config_file = args[1];
      output_file = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java -jar Splitter.jar <file name> <splitter config file> <output file>");
      System.err.println("Tokenizes different chunks of a text file and stores them in an output xml file");
      System.exit(-1);
    }

    try {
      Splitter splitter = new Splitter(config_file);
      splitter.tokenize(input_file);
      splitter.writeDigest(output_file);
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
