package edu.georgetown.gucs.tokenizers;

import java.util.Iterator;

import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;

/**
 * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
 * whitespace
 * 
 * @author Lindsay Neubauer
 * 
 */
public class StripMarkupTokenizer extends Tokenizer {

  private static final long serialVersionUID = -1366304768764987607L;
  private boolean keepScript = false;

  /**
   * Constructor that specifies whether to keep tokens nested inside <code><\script></code> tags; the default is to
   * eliminate these tokens
   * 
   * @param keepScript
   *          the string value (true or false) specifying whether to keep tokens nested inside <code><\script></code>
   *          tags
   */
  public StripMarkupTokenizer(String keepScript) {
    this.keepScript = Boolean.valueOf(keepScript);
  }

  /**
   * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
   * whitespace
   */
  @Override
  public void tokenize(Iterator<String> iterator) {
    synchronized (this.tokenVector) {
      this.tokenVector.clear();
      boolean script = false;
      while (iterator.hasNext()) {
        for (String token : iterator.next().split(">")) {

          // Printing error message instead of throwing exception because govdocs has unbalanced brackets
          if (token.indexOf("<") != token.lastIndexOf("<")) {
            Global.monitor.printErrorMessage("StripMarkupTokenizer", "Unbalanced brackets error: " + token + ">");
            Global.monitor.writeToBuffer("StripMarkupTokenizer", "tokenize", "Unbalanced brackets error: " + token + ">");
          }
          if (token.contains("script") && token.contains("<")) {
            script = !script;
          }

          if (!script) {
            if (token.indexOf('<') > 0) {
              this.tokenVector.add(token.substring(0, token.indexOf('<')));
            } else if (token.indexOf('<') < 0) {
              this.tokenVector.add(token);
            }
          } else if (script && this.keepScript) {
            this.tokenVector.add(token);
          }
        }
      }
    }
  }

  /**
   * Eliminates tokens nested inside markup language tags; assumes that tokens have been split by line rather than using
   * whitespace
   */
  @Override
  public void tokenize(Iterator<String> tokensIterator, Iterator<Pair<Integer, Integer>> positionsIterator) {
    synchronized (this.tokenVector) {
      synchronized (this.positions) {
        this.tokenVector.clear();
        this.positions.clear();
        boolean script = false;
        while (tokensIterator.hasNext() && positionsIterator.hasNext()) {
          Pair<Integer, Integer> pos = positionsIterator.next();
          for (String token : tokensIterator.next().trim().split(">")) {

            // Printing error message instead of throwing exception because govdocs has unbalanced brackets
            if (token.indexOf("<") != token.lastIndexOf("<")) {
              Global.monitor.printErrorMessage("StripMarkupTokenizer", "Unbalanced brackets error: " + token + ">");
              Global.monitor.writeToBuffer("StripMarkupTokenizer", "tokenize_pair", "Unbalanced brackets error: " + token + ">");
            }
            if (token.contains("script") && token.contains("<")) {
              script = !script;
            }

            if (!script) {
              if (token.indexOf('<') > 0) {
                this.tokenVector.add(token.substring(0, token.indexOf('<')));
                this.positions.add(pos);
              } else if (token.indexOf('<') < 0) {
                this.tokenVector.add(token);
                this.positions.add(pos);
              }
            } else if (script && this.keepScript) {
              this.tokenVector.add(token);
              this.positions.add(pos);
            }
          }
        }
      }
    }
  }
}
