package edu.georgetown.gucs.tokenizers;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Pair;

/**
 * Splits a document into tokens (either by line or by word) or alters the tokens in various ways. This is the base
 * class for all tokenizers.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class Tokenizer implements Serializable, Iterable<String> {

  private static final long serialVersionUID = 2475577088718479996L;

  /** The list of each token in order of its appearance */
  protected List<String> tokenVector = Collections.synchronizedList(new Vector<String>());

  /** The list of integer pairs corresponding to the position of each token */
  protected List<Pair<Integer, Integer>> positions = Collections.synchronizedList(new Vector<Pair<Integer, Integer>>());

  /**
   * Provides the list of each token in order of its appearance
   * 
   * @return the list containing each token as a string
   */
  public List<String> getTokenVector() {
    synchronized (this.tokenVector) {
      return this.tokenVector;
    }
  }

  /**
   * Provides the list of integer pairs corresponding to the position of each token
   * 
   * @return the list of integer pairs
   */
  public List<Pair<Integer, Integer>> getPositionsVector() {
    synchronized (this.positions) {
      return this.positions;
    }
  }

  /**
   * Alters or eliminates certain tokens. By default, this method adds all string tokens from the given iterator to the
   * list of tokens, and must be over-ridden in each specific tokenizer.
   * 
   * @param iterator
   *          the string iterator over the token elements
   */
  public void tokenize(Iterator<String> iterator) {
    synchronized (this.tokenVector) {
      this.tokenVector.clear();
      while (iterator.hasNext()) {
        this.tokenVector.add(iterator.next().trim());
      }
    }
  }

  /**
   * Alters or eliminates certain tokens when using Splitter. By default, this method adds all string tokens and pairs
   * from the given iterators to the lists of tokens and positions, respectively, and must be over-ridden in each
   * specific tokenizer.
   * 
   * @param tokensIterator
   *          the string iterator over the token elements
   * @param positionsIterator
   *          the integer pair iterator over the start and end position elements
   */
  public void tokenize(Iterator<String> tokensIterator, Iterator<Pair<Integer, Integer>> positionsIterator) {
    synchronized (this.tokenVector) {
      synchronized (this.positions) {
        this.tokenVector.clear();
        this.positions.clear();
        while (tokensIterator.hasNext() && positionsIterator.hasNext()) {
          this.tokenVector.add(tokensIterator.next().trim());
          this.positions.add(positionsIterator.next());
        }
      }
    }
  }

  /**
   * Splits the document into tokens. By default, this method calls the ExceptionHandler to throw a
   * "Tokenizer does not support reading from a file" exception. It must be over-ridden in each tokenizer that supports
   * reading from a file.
   * 
   * @param filename
   *          the filename of the document to split into tokens
   */
  public void tokenize(String filename) {
    ExceptionHandler.handleExceptionThrow(new Exception("Tokenizer does not support reading from a file"));
  }

  /**
   * Provides an iterator over the tokens in this tokenizer
   * 
   * @return the iterator over the string tokens
   */
  @Override
  public Iterator<String> iterator() {
    synchronized (this.tokenVector) {
      return this.tokenVector.iterator();
    }
  }

  /**
   * Provides an iterator over the positions of elements in this tokenizer
   * 
   * @return the iterator over the integer pairs of start and end position elements
   */
  public Iterator<Pair<Integer, Integer>> position_iterator() {
    synchronized (this.positions) {
      return this.positions.iterator();
    }
  }

  /**
   * Prints each token to the system output stream using the UTF-8 charset
   * 
   * @param tokens
   *          the string iterator over the tokens in this tokenizer
   */
  public static void printTokens(Iterator<String> tokens) {
    try {
      // Not sure this "UTF-8" does anything
      OutputStreamWriter out = new OutputStreamWriter(System.out, "UTF-8");
      BufferedWriter writer = new BufferedWriter(out);
      while (tokens.hasNext()) {
        writer.write(String.format("%s%n", tokens.next()));
      }
      out.flush();
      writer.flush();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
  }
}
