
/*
| Copyright (c) 2001, 2011, Oracle and/or its affiliates. All rights reserved.
|
| This file contains material copyrighted by Oracle and is the
| confidential unpublished property of Oracle, and may not be
| copied in whole or part, or disclosed to any third party, without the
| written prior consent of Oracle.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "scctype.h"
#include "sccca.h"

//  The size of the buffer to be filled with CA textual content.
//  This buffer must be at least 1K bytes and should be 2 byte aligned.
#define CONTENT_BUFFER_SIZE  0x1000

// Used with DAOpenDocument to indicate how a document is being specified.
#ifdef UNIX
#define PATH_TYPE   IOTYPE_UNIXPATH
#else
#define PATH_TYPE   IOTYPE_ANSIPATH
#endif

void runTextExtraction(const char* filename);
void printErrorType(char* method, char* filename, DAERR daErr);
void setOptions(VTHDOC const hDoc, char* filename);

/** 
 * Extract text from document
 *
 * @param const char* filename   file name to extract text
**/
void runTextExtraction(const char* filename) {
    VTHDOC hDoc;
    DAERR daErr;
    VTLONG threadErr;
    VTDWORD optionValue;
    VTHCONTENT hcaContent;
    SCCCAGETCONTENT caContent;
    VTBYTE buffer[CONTENT_BUFFER_SIZE];
    
    // Initialize the Data Access module
    daErr = DAInit();
    if (daErr != DAERR_OK) {
        printErrorType("DAInit", filename, daErr);
        return NULL;
    }

    // Open input document
    optionValue = SCCUT_FI_EXTENDEDTEST;
    daErr = DASetOption(0, SCCOPT_FIFLAGS, &optionValue, sizeof(optionValue));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_FIFLAGS", filename, daErr);
        DADeInit();
        return NULL;
    }
    
    // DAOpenDocument creates a handle to a document
    daErr = DAOpenDocument(&hDoc, PATH_TYPE, filename, 0);
    if (daErr != DAERR_OK) {
        printErrorType("DAOpenDocument", filename, daErr);
        DADeInit();
        return NULL;
    }

    // Set the document specific options
    setOptions(hDoc, filename);

    // Initialize the CA content handle
    daErr = CAOpenContent(hDoc, &hcaContent);
    if (daErr != DAERR_OK) {
        printErrorType("CAOpenContent", filename, daErr);
        DACloseDocument(hDoc);
        DADeInit();
        return NULL;
    }   

    // Set up the structure used to return CA items
    caContent.dwStructSize = sizeof(SCCCAGETCONTENT);
    caContent.dwFlags      = 0;
    caContent.dwMaxBufSize = CONTENT_BUFFER_SIZE;
    caContent.pDataBuf     = &buffer;

    // Read first CA item
    daErr = CAReadFirst(hcaContent, &caContent);
    if (daErr != DAERR_OK) {
        printErrorType("CAReadFirst", filename, daErr);
        CACloseContent(hcaContent);
        DACloseDocument(hDoc);
        DADeInit();
        return NULL;
    }
    
    // Process CA items
    while (daErr == DAERR_OK || daErr == DAERR_MISALIGNMENT) {
        if (daErr == DAERR_MISALIGNMENT) {
          printf("*** A misalignment error was generated for type 0x%08x ***\n", caContent.dwType);
        } else if (daErr == DAERR_OK) {
          fprintf(stdout, &buffer, " ");
          memset(&buffer, 0, sizeof(buffer));
        } else {
          printf("\t***  Error:  Unexpected type (0x%08x) from CA.  ***\n", caContent.dwType);
        }
        daErr = CAReadNext(hcaContent, &caContent);
    }
    
    // EOF is normal successful completion
    if (daErr != DAERR_EOF) {
      printErrorType("CAReadNext", filename, daErr);
      CACloseContent(hcaContent);
      DACloseDocument(hDoc);
      DADeInit();
      return NULL;
    }
    
    // Close the CA content
    daErr = CACloseContent(hcaContent);
    if (daErr != DAERR_OK) {
        printErrorType("CACloseContent", filename, daErr);
    }
    
    // Close the document
    daErr = DACloseDocument(hDoc);
    if (daErr != DAERR_OK) {
        printErrorType("DACloseDocument", filename, daErr);
        DADeInit();
        return NULL;
    }

    // Deinitialize DA
    daErr = DADeInit();
    if (daErr != DAERR_OK) {
        printErrorType("DADeInit", filename, daErr);
    }
    
}

/**
 * Prints error message
 *
 * @param char* method   method name where error occurred
 * @param char* file     file name where errror occurred
 * @param DAERR daErr    type of error
**/
void printErrorType(char* method, char* file, DAERR daErr) {
    char* error = "";

    if (daErr == DAERR_BADPARAM)  error = "(DAERR_BADPARAM)";
    else if (daErr == DAERR_EMPTYFILE) error = "(DAERR_EMPTYFILE)";
    else if (daErr == DAERR_PROTECTEDFILE) error = "(DAERR_PROTECTEDFILE)";
    else if (daErr == DAERR_SUPFILEOPENFAILS)  error = "(DAERR_SUPFILEOPENFAILS)";
    else if (daErr == DAERR_FILTERNOTAVAIL)  error = "(DAERR_FILTERNOTAVAIL)";
    else if (daErr == DAERR_FILTERLOADFAILED)  error = "(DAERR_FILTERLOADFAILED)";
    else if (daErr == DAERR_BADFILE) error = "(DAERR_BADFILE)";
    else if (daErr == DAERR_DRMFILE) error = "(DAERR_DRMFILE)";
    else if (daErr == DAERR_MEMORY)  error = "(DAERR_MEMORY)";
    else if (daErr == DAERR_ABORT) error = "(DAERR_ABORT)";
    else if (daErr == DAERR_FILEOPENFAILED)  error = "(DAERR_FILEOPENFAILED)";
    else if (daErr == DAERR_CHARMAPFAILED) error = "(DAERR_CHARMAPFAILED)";
    else if (daErr == DAERR_UNKNOWN) error = "(DAERR_UNKNOWN)";
    else if (daErr == DAERR_FILECREATE)  error = "(DAERR_FILECREATE)";
    else if (daErr == DAERR_FILEWRITEFAILED) error = "(DAERR_FILEWRITEFAILED)";
    else if (daErr == DAERR_MORE)  error = "(DAERR_MORE)";
    else if (daErr == DAERR_NOFILE)  error = "(DAERR_NOFILE)";
    else if (daErr == DAERR_UNSUPPORTEDCOMPRESSION)  error = "(DAERR_UNSUPPORTEDCOMPRESSION)";
    else if (daErr == DAERR_MISALIGNMENT)  error = "(DAERR_MISALIGNMENT)";
    else if (daErr == DAERR_NODATATORENDER)  error = "(DAERR_NODATATORENDER)";
    else if (daErr == DAERR_EXTERNALURLREFERENCE)  error = "(DAERR_EXTERNALURLREFERENCE)";
    else if (daErr == DAERR_LINKTOFILE)  error = "(DAERR_LINKTOFILE)";
    else if (daErr == DAERR_NOTARCHIVEFILE)  error = "(DAERR_NOTARCHIVEFILE)";
    else if (daErr == DAERR_NOTSUPPORTEDARCHIVEFILE) error = "(DAERR_NOTSUPPORTEDARCHIVEFILE)";
    else if (daErr == DAERR_BADCREDENTIALS)  error = "(DAERR_BADCREDENTIALS)";
    
    fprintf (stderr, "\nERROR: %s %s %s\n", method, error, file);
}

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

/**
 * Sets document specific options for Content Access 
 * This method is very similar to Oracle's sample code in casample.c
 *
 * @param VTHDOC const hDoc    document handle to set options
 * @param char* filename       file name of document to sent options
**/
void setOptions(VTHDOC const hDoc, char* filename) {
    DAERR daErr;
    VTDWORD dwordOptionValue;
    VTBOOL  bExtractXMPMetadata;

    //  Turn on processing of generated text
    dwordOptionValue = SCCEX_IND_GENERATED;
    daErr = DASetOption(hDoc, SCCOPT_CA_FLAGS, &dwordOptionValue, sizeof(dwordOptionValue));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_CA_FLAGS", filename, daErr);
    }

    //  Set the fallback format
    dwordOptionValue = FI_TEXT;
    daErr = DASetOption(hDoc, SCCOPT_FALLBACKFORMAT, &dwordOptionValue, sizeof(dwordOptionValue));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_FALLBACKFORMAT", filename, daErr);
    }

    //  Set default input character set
    dwordOptionValue = SO_ANSI1252;
    daErr = DASetOption(hDoc, SCCOPT_DEFAULTINPUTCHARSET, &dwordOptionValue, sizeof(dwordOptionValue));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_DEFAULTINPUTCHARSET", filename, daErr);
    }

    //  Set output character set
    dwordOptionValue = SO_ANSI1252;
    daErr = DASetOption(hDoc, SCCOPT_OUTPUTCHARACTERSET, &dwordOptionValue, sizeof(dwordOptionValue));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_OUTPUTCHARACTERSET", filename, daErr);
    }

    //  File Identification code will run an extended test on all files that are not identified
    dwordOptionValue = SCCUT_FI_EXTENDEDTEST;
    daErr = DASetOption(hDoc, SCCOPT_FIFLAGS, &dwordOptionValue, sizeof(dwordOptionValue));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_FIFLAGS", filename, daErr);
    }

    // set extract xmp metadata
    bExtractXMPMetadata = FALSE;
    //bExtractXMPMetadata = TRUE;
    daErr = DASetOption(hDoc, SCCOPT_EXTRACTXMPMETADATA, &bExtractXMPMetadata, sizeof(VTBOOL));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_EXTRACTXMPMETADATA", filename, daErr);
    }

    // Ignore validation of the password
    bExtractXMPMetadata = TRUE;
    daErr = DASetOption(hDoc, SCCOPT_IGNORE_PASSWORD, &bExtractXMPMetadata, sizeof(VTBOOL));
    if (daErr != DAERR_OK) {
        printErrorType("DASetOption SCCOPT_IGNORE_PASSWORD", filename, daErr);
    }
}

