package edu.georgetown.gucs.utility;

import java.io.File;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * Creates a list of files in a directory and allows iteration over all and parts of this list
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class FileLister implements Serializable, Iterable<File> {

  private static final long serialVersionUID = -8536944759634384952L;
  private File directory;
  private int maxThread;
  private LinkedBlockingQueue<File> fileQueue = new LinkedBlockingQueue<File>();
  private List<File> allFiles = new Vector<File>();
  private List<File> dictionaryList = new Vector<File>();
  private List<File> sampleList = new Vector<File>();
  private Map<String, Integer> fileTypes = new HashMap<String, Integer>();
  private boolean loaded = false;

  /**
   * Constructor that sets the maximum thread count to the number of processors available to the Java virtual machine
   */
  public FileLister() {
    this.maxThread = Runtime.getRuntime().availableProcessors();
    Global.monitor.writeToBuffer("FileLister", "Constructor", "maxThread=" + this.maxThread);
  }

  /**
   * Constructor that recursively loads files from the given directory and sets the maximum thread count to the number
   * of processors available to the Java virtual machine
   * 
   * @param directory
   *          the string path to the directory to load
   */
  public FileLister(String directory) {
    this.directory = new File(directory);
    this.maxThread = Runtime.getRuntime().availableProcessors();
    Global.monitor.writeToBuffer("FileLister", "Constructor", "maxThread=" + this.maxThread);
    recursiveLoadDirectory(directory);
  }

  /**
   * Constructor that recursively loads files from the given directory and sets the maximum thread count
   * 
   * @param directory
   *          the string path to the directory to load
   * @param maxThread
   *          the <code>int</code> maximum threads to use to run this fileLister
   */
  public FileLister(String directory, int maxThread) {
    this.directory = new File(directory);
    this.maxThread = maxThread;
    Global.monitor.writeToBuffer("FileLister", "Constructor", "maxThread=" + maxThread);
    recursiveLoadDirectory(directory);
  }

  /**
   * Constructor that recursively loads files from the given directory, sets the maximum thread count, and loads
   * separate dictionary and sample lists of the given sizes; dictionary and sample lists do not share any files
   * 
   * @param directory
   *          the string path to the directory to load
   * @param maxThread
   *          the <code>int</code> maximum threads to use to run this fileLister
   * @param dictionarySize
   *          the <code>double</code> number of files to include in the dictionary list
   * @param sampleSize
   *          the <code>double</code> number of files to include in the sample list
   */
  public FileLister(String directory, int maxThread, double dictionarySize, double sampleSize) {
    this.directory = new File(directory);
    this.maxThread = maxThread;
    Global.monitor.writeToBuffer("FileLister", "Constructor", "maxThread=" + maxThread);
    recursiveLoadDirectory(directory);
    createSublists(dictionarySize, sampleSize);
  }

  /**
   * Populates the dictionary and sample lists, waiting for all files to be loaded before randomly assigning files to
   * one of the lists. Dictionary and sample lists do not share any files. If this method is never called all files are
   * assigned to the dictionary list.
   * 
   * @param dictionarySize
   *          the <code>double</code> number of files to include in the dictionary list
   * @param sampleSize
   *          the <code>double</code> number of files to include in the sample list
   */
  public void createSublists(double dictionarySize, double sampleSize) {
    if (dictionarySize < 0 || sampleSize < 0) {
      ExceptionHandler.handleExceptionThrow(new Exception("FileLister: Negative size parameter(s): dictionarySize = "
          + dictionarySize + "  sampleSize = " + sampleSize));
    }
    int dictionaryCount = (int) ((dictionarySize >= 1) ? dictionarySize
        : ((dictionarySize * 10 * this.allFiles.size()) + .5));
    int sampleCount = (int) ((sampleSize >= 1) ? sampleSize : ((sampleSize * 10 * this.allFiles.size()) + .5));
    Global.monitor.printMessage("FileLister", "createSublists: dictionaryCount=" + dictionaryCount + "  sampleCount="
        + sampleCount);
    Global.monitor.writeToBuffer("FileLister", "createSublists", "dictionaryCount=" + dictionaryCount);
    Global.monitor.writeToBuffer("FileLister", "createSublists", "sampleCount=" + sampleCount);

    boolean loop = true;
    while (loop) {
      if (this.loaded) {
        loop = false;
        if (dictionaryCount > this.allFiles.size() - sampleCount
            || sampleCount > this.allFiles.size() - dictionaryCount) {
          ExceptionHandler.handleExceptionThrow(new Exception(String.format("%s%n  %s%n  %s%n  %s%n", 
              "FileLister: Incompatible parameters: dictionary files + sample files <= total directory files", 
              "Total directory files: "+ this.allFiles.size(),
              "Requested number of dictionary files: " + dictionaryCount,
              "Requested number of sample files: " + sampleCount)));
        }

        Collections.shuffle(this.allFiles);
        this.dictionaryList.clear();
        this.sampleList.clear();
        this.dictionaryList.addAll(this.allFiles.subList(0, dictionaryCount));
        this.sampleList.addAll(this.allFiles.subList(dictionaryCount, dictionaryCount + sampleCount));
        Global.monitor.printMessage("FileLister", "createSublists: dictionaryList.size=" + this.dictionaryList.size() + "  sampleList.size="
            + this.sampleList.size());
      }
    }
    Global.monitor.printMessage("FileLister", "createSublists: completed");

  }

  /**
   * Provides the counts for each file extension that exists in this list of files
   * 
   * @return the hashMap of file extensions and their integer counts in this list of files
   */
  public Map<String, Integer> getFileTypes() {
    return this.fileTypes;
  }

  /**
   * Provides an iterator over the list of all files
   * 
   * @return the file iterator over the list of all files
   */
  @Override
  public Iterator<File> iterator() {
    return this.allFiles.iterator();
  }

  /**
   * Provides an iterator over the list of files used for a sample
   * 
   * @return the file iterator over the list of files used for a sample
   */
  public Iterator<File> sample_iterator() {
    return this.sampleList.iterator();
  }

  /**
   * Provides an iterator over the list of files used for a dictionary
   * 
   * @return the file iterator over the list of files used for a dictionary
   */
  public Iterator<File> dictionary_iterator() {
    return this.dictionaryList.iterator();
  }

  /**
   * Provides the size of the list of all files
   * 
   * @return the <code>int</code> number of files
   */
  public int size() {
    return this.allFiles.size();
  }
  
  /**
   * Provides the size of the list of dictionary files
   * 
   * @return the <code>int</code> number of files used for the dictionary
   */
  public int dictionary_size() {
    return this.dictionaryList.size();
  }
  
  /**
   * Provides the size of the list of files used for the sample
   * 
   * @return the <code>int</code> number of files in the sample list
   */
  public List<File> getSampleFiles() {
    return this.sampleList;
  }

  /**
   * Determines whether all the files have been loaded and unloaded from the queue of files
   * 
   * @return the <code>boolean</code> value whether all the files have been loaded and unloaded from the queue of files
   */
  public synchronized boolean loadComplete() {
    return (this.loaded && this.fileQueue.size() == 0);
  }

  /**
   * Provides the path to the directory used for the initial list of files
   * 
   * @return the string path to the directory used to initially create this list of files
   */
  public String getDirectory() {
    return this.directory.getAbsolutePath();
  }

  /**
   * Retrieves and removes the head of the file queue, waiting if necessary until an element becomes available
   * 
   * @return the file at the head of the file queue
   */
  public synchronized File takeFile() {
    try {
      return this.fileQueue.take();
    } catch (InterruptedException e) {
      Global.monitor.writeToBuffer("FileLister", "takeFile", "thread_id = " + Thread.currentThread().getId());
      Global.monitor.writeToBuffer("FileLister", "takeFile", Arrays.toString(Thread.currentThread().getStackTrace()));
      ExceptionHandler.handleExceptionExit(e);
    }
    return null;
  }

  /**
   * Adds the file to the tail of the file queue, waiting if necessary for space to become available
   * 
   * @param file
   *          the file to put at the tail of the file queue
   */
  protected synchronized void putFile(File file) {
    try {
      this.fileQueue.put(file);
    } catch (InterruptedException e) {
      Global.monitor.writeToBuffer("FileLister", "putFile", "thread_id = " + Thread.currentThread().getId());
      Global.monitor.writeToBuffer("FileLister", "putFile", Arrays.toString(Thread.currentThread().getStackTrace()));
      ExceptionHandler.handleExceptionExit(e);
    }
  }

  /**
   * Adds the file to the end of this list of files used for a dictionary
   * 
   * @param file
   *          the file to add to this list of files
   */
  protected void addFile(File file) {
    this.allFiles.add(file);
    this.dictionaryList.add(file);
  }

  /**
   * Loads all the files from the given directory into the list of files used for a dictionary
   * 
   * @param path
   *          the string path to the directory to load
   */
  public void loadDirectory(String path) {
    Global.monitor.writeToBuffer("FileLister", "loadDirectory", "directory=" + path);

    this.directory = new File(path);
    ExecutorService exec = Executors.newFixedThreadPool(this.maxThread);
    BoundedExecutor bex = new BoundedExecutor(exec, this.maxThread);
    for (File file : Arrays.asList(this.directory.listFiles())) {
      if (file.isFile() && file.canRead()) {
        FileListerWorker currentFile = new FileListerWorker(file, this);
        bex.submitTask(currentFile);
      }
    }

    exec.shutdown();
    while (!exec.isTerminated()) {
      try {
        // Wait a second and see if we are done
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        Global.monitor.writeToBuffer("FileLister", "loadDirectory", "thread_id = " + Thread.currentThread().getId());
        Global.monitor.writeToBuffer("FileLister", "loadDirectory", Arrays.toString(Thread.currentThread().getStackTrace()));
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    this.loaded = true;
    Global.monitor.printMessage("FileLister", "Files Loaded: " + this.directory.getAbsolutePath());
    Global.monitor.writeToBuffer("FileLister", "loadDirectory", "loaded=" + this.loaded);
  }

  /**
   * Recursively loads all the files from the given directory into the list of files used for a dictionary
   * 
   * @param path
   *          the string path to the directory to load
   */
  public void recursiveLoadDirectory(String path) {
    Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "directory=" + path);

    File root = new File(path);
    ExecutorService exec = Executors.newFixedThreadPool(this.maxThread);
    BoundedExecutor bex = new BoundedExecutor(exec, this.maxThread);
    for (File file : Arrays.asList(root.listFiles())) {
      if (file.isFile() && file.canRead()) {
        FileListerWorker currentFile = new FileListerWorker(file, this);
        bex.submitTask(currentFile);
      } else if (file.isDirectory() && file.canRead()) {
        // Ignore .svn directories
        if (file.toString().indexOf(".svn") < 0) {
          recursiveLoadDirectory(file.toString());
        }
      }
    }

    exec.shutdown();
    while (!exec.isTerminated()) {
      try {
        // Wait a second and see if we are done
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "thread_id = " + Thread.currentThread().getId());
        Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", Arrays.toString(Thread.currentThread().getStackTrace()));
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    this.loaded = true;
    Global.monitor.printMessage("FileLister", "Files Loaded: " + root.getAbsolutePath());
    Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "loaded=" + this.loaded);
  }

  /**
   * Returns a list of each of the files in this fileLister, each on a separate line
   */
  @Override
  public String toString() {
    StringBuffer string = new StringBuffer();
    for (File file : this.allFiles) {
      string.append(String.format("%s%n", file.toString()));
    }
    return string.toString();
  }

  /**
   * Increments the count of the file extension type for the given filename
   * 
   * @param filename
   *          the string filename of file extension to add
   */
  protected void updateFileTypes(String filename) {
    String extension = filename.substring(filename.lastIndexOf('.') + 1).toUpperCase(Locale.ENGLISH);

    if (this.fileTypes.containsKey(extension)) {
      int count = this.fileTypes.get(extension);
      this.fileTypes.put(extension, count + 1);
    } else {
      this.fileTypes.put(extension, 1);
    }
  }
}
